import numpy as np
import matplotlib.pyplot as plt

from amisc.component import SparseGridSurrogate
from amisc.rv import UniformRV
from amisc.utils import ax_default


def test_sparse_grid(plots=False):
    """Simple cos test from Jakeman (2022)"""
    def model(x, alpha):
        alpha = np.atleast_1d(alpha)  # (1,)
        eps = (1/5) * 2.0**(-alpha[0])
        y = np.cos(np.pi/2 * (x + 4/5 + eps))
        return {'y': y}

    # Construct MISC surrogate from an index set
    Ik = [((0,), (0,)), ((0,), (1,)), ((1,), (0,)), ((2,), (0,)), ((1,), (1,)), ((0,), (2,)), ((1,), (2,)),
          ((2,), (1,)), ((2,), (2,))]
    x_vars = UniformRV(-1, 1)
    truth_alpha = (15,)
    comp = SparseGridSurrogate(x_vars, model, multi_index=Ik, truth_alpha=truth_alpha)
    N = 100
    xg = np.linspace(-1, 1, N).reshape((N, 1))
    yt = comp(xg, use_model=truth_alpha)
    y_surr = comp(xg)
    l2_error = np.sqrt(np.mean((y_surr - yt)**2)) / np.sqrt(np.mean(yt**2))
    assert l2_error < 0.1

    # Plot results for each fidelity of the MISC surrogate
    if plots:
        fig, axs = plt.subplots(3, 3, sharey='row', sharex='col')
        for alpha in range(3):
            for beta in range(3):
                ax = axs[2-alpha, beta]
                surr = comp.get_sub_surrogate((alpha,), (beta,), include_grid=True)
                s = f'$\hat{{f}}_{{{alpha}, {beta}}}$'
                ax.plot(xg, surr(xg), '--k', label=r'{}'.format(s), linewidth=1.5)
                s = f'$\hat{{f}}_{alpha}$'
                ax.plot(xg, model(xg, alpha)['y'], '--b', label=r'{}'.format(s), linewidth=2)
                ax.plot(xg, yt, '-r', label=r'$f$', linewidth=2)
                ax.plot(surr.xi, surr.yi, 'or')
                xlabel = r'$x$' if alpha == 0 else ''
                ylabel = r'$f(x)$' if beta == 0 else ''
                ax_default(ax, xlabel, ylabel, legend=True)

        fig.text(0.5, 0.02, r'Increasing surrogate fidelity ($\beta$) $\rightarrow$', ha='center', fontweight='bold')
        fig.text(0.02, 0.5, r'Increasing model fidelity ($\alpha$) $\rightarrow$', va='center', fontweight='bold', rotation='vertical')
        fig.set_size_inches(3 * 3, 3 * 3)
        fig.tight_layout(pad=3, w_pad=1, h_pad=1)
        plt.show()

        fig, ax = plt.subplots()
        ax.plot(xg, yt, '-r', linewidth=2, label='Model')
        ax.plot(xg, y_surr, '--k', linewidth=1.5, label='MISC surrogate')
        ax_default(ax, r'$x$', r'$f(x)$', legend=True)
        plt.show()


# TODO: add tests for testing parallel execution, writing output files, and refining surrogate



"""`interpolator.py`

Provides interpolator classes. Interpolators manage training data and specify how to refine/gather more data.

Includes
--------
- `BaseInterpolator`: Abstract class providing basic structure of an interpolator
- `LagrangeInterpolator`: Concrete implementation for tensor-product barycentric Lagrange interpolation
"""
from abc import ABC, abstractmethod
import itertools
import copy

import numpy as np
from scipy.optimize import direct
from sklearn.linear_model import Ridge
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import MaxAbsScaler

from amisc.utils import get_logger
from amisc.rv import BaseRV


class BaseInterpolator(ABC):
    """Base interpolator abstract class.

    !!! Info "Setting the training data"
        You can leave the training data `xi`, `yi` empty; they can be iteratively refined later on.

    !!! Info "Model specification"
        The model is a callable function of the form `ret = model(x, *args, **kwargs)`. The return value is a dictionary
        of the form `ret = {'y': y, 'files': files, 'cost': cost}`. In the return dictionary, you specify the raw model
        output `y` as an `np.ndarray` at a _minimum_. Optionally, you can specify paths to output files and the average
        model cost (in units of seconds of cpu time), and anything else you want.

    :ivar beta: specifies the refinement level of this interpolator as a set of natural number indices
    :ivar x_vars: list of variables that fully determines the input domain of interest for interpolation
    :ivar xi: `(Nx, x_dim)`, interpolation points (or knots, training samples, etc.) stored as an array
    :ivar yi: `(Nx, y_dim)`, function values at the interpolation points, i.e. the training data
    :ivar _model: stores a ref to the model or function that is to be interpolated, callable as `ret = model(x)`
    :ivar _model_args: additional arguments to supply to the model
    :ivar _model_kwargs: additional keyword arguments to supply to the model
    :ivar model_cost: the average total cpu time (in seconds) for a single model evaluation call of one set of inputs
    :ivar output_files: tracks model output files corresponding to `yi` training data (for more complex models)
    :ivar logger: a logging utility reference

    :vartype beta: tuple[int, ...]
    :vartype x_vars: list[BaseRV]
    :vartype xi: np.ndarray
    :vartype yi: np.ndarray
    :vartype _model: callable[np.ndarray] -> dict
    :vartype _model_args: tuple
    :vartype _model_kwargs: dict
    :vartype model_cost: float
    :vartype output_files: list[str | Path]
    :vartype logger: logging.Logger
    """

    def __init__(self, beta: tuple, x_vars: BaseRV | list[BaseRV], xi=None, yi=None,
                 model=None, model_args=(), model_kwargs=None):
        """Construct the interpolator.

        :param beta: refinement level indices
        :param x_vars: list of variables to specify input domain of interpolation
        :param xi: `(Nx, xdim)`, interpolation points (optional)
        :param yi: `(Nx, ydim)`, the function values at the interpolation points (optional)
        :param model: callable as {'y': y} = model(x), with `x = (..., x_dim)`, `y = (..., y_dim)`
        :param model_args: optional args for the model
        :param model_kwargs: optional kwargs for the model
        """
        x_vars = [x_vars] if not isinstance(x_vars, list) else x_vars
        self.logger = get_logger(self.__class__.__name__)
        self._model = model
        self._model_args = model_args
        self._model_kwargs = model_kwargs if model_kwargs is not None else {}
        self.output_files = []                              # Save output files with same indexing as xi, yi
        self.xi = xi                                        # Interpolation points
        self.yi = yi                                        # Function values at interpolation points
        self.beta = beta                                    # Refinement level indices
        self.x_vars = x_vars                                # BaseRV() objects for each input
        self.model_cost = None                              # Total cpu time to evaluate model once (s)

    def update_input_bds(self, idx: int, bds: tuple):
        """Update the input bounds at the given index.

        :param idx: the index of the input variable to update
        :param bds: the new bounds for the variable
        """
        self.x_vars[idx].update_bounds(*bds)

    def xdim(self):
        """Get the dimension of the input domain."""
        return len(self.x_vars)

    def ydim(self):
        """Get the dimension of the outputs."""
        return self.yi.shape[-1] if self.yi is not None else None

    def save_enabled(self):
        """Return whether the underlying model wants to save outputs to file.

        !!! Note
            You can specify that a model wants to save outputs to file by providing an `'output_dir'` kwarg.
        """
        return self._model_kwargs.get('output_dir') is not None

    def set_yi(self, yi: np.ndarray = None, model: callable = None,
               x_new: tuple[list[int | tuple], np.ndarray] = ()) -> dict[str: np.ndarray] | None:
        """Set the training data; if `yi=None`, then compute from the model.

        !!! Warning
            You would use `x_new` if you wanted to compute the model at these specific locations and store the result.
            This will ignore anything passed in for `yi`, and it assumes a model is already specified (or passed in).

        !!! Info
            You can pass in integer indices for `x_new` or tuple indices. Integers will index into `self.xi`. Tuples
            provide extra flexibility for more complicated indexing, e.g. they might specify indices along different
            coordinate directions in an N-dimensional grid. If you pass in a list of tuple indices for `x_new`, the
            resulting model outputs will be returned back to you in the form `dict[str: np.ndarray]`. The keys are
            string casts of the tuple indices, and the values are the corresponding model outputs.

        :param yi: `(Nx, y_dim)`, training data to set, must match dimension of `self.xi`
        :param model: callable function, optionally overrides `self._model`
        :param x_new: tuple of `(idx, x)`, where `x` is an `(N_new, x_dim)` array of new interpolation points to
                      include and `idx` specifies the indices of these points into `self.xi`
        :returns: dict[str: np.ndarray] if `idx` contains tuple elements, otherwise `None`
        """
        if model is not None:
            self._model = model
        if self._model is None:
            error_msg = 'Model not specified for computing QoIs at interpolation grid points.'
            self.logger.error(error_msg)
            raise Exception(error_msg)

        # Overrides anything passed in for yi (you would only be using this if yi was set previously)
        if x_new:
            new_idx = x_new[0]
            new_x = x_new[1]
            return_y = isinstance(new_idx[0], tuple)  # Return y rather than storing it if tuple indices are passed in
            ret = dict(y=dict(), files=dict())
            model_ret = self._model(new_x, *self._model_args, **self._model_kwargs)
            if not isinstance(model_ret, dict):
                self.logger.warning(
                    f"Function {self._model} did not return a dict of the form {{'y': y}}. Please make sure"
                    f" you do so to avoid conflicts. Returning the value directly instead...")
                model_ret = dict(y=model_ret)
            y_new, files_new, cpu_time = model_ret['y'], model_ret.get('files', None), model_ret.get('cost', 1)

            if self.save_enabled():
                for j in range(y_new.shape[0]):
                    if return_y:
                        ret['y'][str(new_idx[j])] = y_new[j, :].astype(np.float32)
                        ret['files'][str(new_idx[j])] = files_new[j]
                    else:
                        self.yi[new_idx[j], :] = y_new[j, :].astype(np.float32)
                        self.output_files[new_idx[j]] = files_new[j]
            else:
                for j in range(y_new.shape[0]):
                    if return_y:
                        ret['y'][str(new_idx[j])] = y_new[j, :].astype(np.float32)
                    else:
                        self.yi[new_idx[j], :] = y_new[j, :].astype(np.float32)

            if self.model_cost is None:
                self.model_cost = max(1, cpu_time)

            return ret

        # Set yi directly
        if yi is not None:
            self.yi = yi.astype(np.float32)
            return

        # Compute yi
        model_ret = self._model(self.xi, *self._model_args, **self._model_kwargs)
        if not isinstance(model_ret, dict):
            self.logger.warning(f"Function {self._model} did not return a dict of the form {{'y': y}}. Please make sure"
                                f" you do so to avoid conflicts. Returning the value directly instead...")
            model_ret = dict(y=model_ret)

        self.yi, self.output_files, cpu_time = model_ret['y'], model_ret.get('files', list()), model_ret.get('cost', 1)

        if self.model_cost is None:
            self.model_cost = max(1, cpu_time)

    @abstractmethod
    def refine(self, beta: tuple, auto=True):
        """Return a new interpolator with one dimension refined by one level, as specified by `beta`.

        !!! Info "When you want to compute the model manually"
            You can set `auto=False`, in which case the newly refined interpolation points `x` will be returned to you
            along with their indices, in the form `idx, x, interp = refine(beta, auto=False)`. You might also want to
            do this if you did not provide a model when constructing the Interpolator (so `auto=True` won't work).

        :param beta: the new refinement level indices, should only refine one dimension by one level
        :param auto: whether to automatically compute and store model at refinement points (default is True)
        :returns: `idx` - indices into `xi`, `x` - the new interpolation points, and `interp` - a refined
                   BaseInterpolator object, just returns `interp` if `auto=True`
        """
        pass

    @abstractmethod
    def __call__(self, x: np.ndarray | float) -> np.ndarray:
        """Evaluate the interpolation at points `x`.

        :param x: `(..., x_dim)`, the points to be interpolated, must be within the input domain for accuracy
        :returns y: `(..., y_dim)`, the interpolated function values
        """
        pass


class LagrangeInterpolator(BaseInterpolator):
    """Tensor-product (multivariate) grid interpolator, based on barycentric Lagrange polynomials.

    !!! Info
        The refinement level indices `beta` are used in this class to specify anisotropic refinement along each
        coordinate direction of the input domain, so `x_dim = len(beta)`.

    :ivar x_grids: univariate Leja sequence points in each 1d dimension
    :ivar weights: the barycentric weights corresponding to `x_grids`
    :ivar reduced: whether to store `xi` and `yi` training data, can set to `False` to save memory, e.g. if an external
                   sparse grid data structure manages this data instead

    :vartype x_grids: list[np.ndarray]
    :vartype weights: list[np.ndarray]
    :vartype reduced: bool
    """

    def __init__(self, beta: tuple, x_vars: BaseRV | list[BaseRV], init_grids=True, reduced=False, **kwargs):
        """Initialize a Lagrange tensor-product grid interpolator.

        :param beta: refinement level indices for each input dimension
        :param x_vars: list of variables specifying bounds/pdfs for each input x
        :param init_grids: whether to compute 1d Leja sequences on initialization
        :param reduced: whether to store xi/yi matrices, e.g. set true if storing in external sparse grid structure
        :param **kwargs: other optional arguments (see `BaseInterpolator`)
        """
        self.weights = []   # Barycentric weights for each dimension
        self.x_grids = []   # Univariate nested leja sequences in each dimension
        self.reduced = reduced
        super().__init__(beta, x_vars, **kwargs)

        if init_grids:
            # Construct 1d univariate Leja sequences in each dimension
            grid_sizes = self.get_grid_sizes(self.beta)
            self.x_grids = [self.leja_1d(grid_sizes[n], self.x_vars[n].bounds(),
                                         wt_fcn=self.x_vars[n].pdf).astype(np.float32) for n in range(self.xdim())]

            for n in range(self.xdim()):
                Nx = grid_sizes[n]
                bds = self.x_vars[n].bounds()
                grid = self.x_grids[n]
                C = (bds[1] - bds[0]) / 4.0  # Interval capacity (see Berrut and Trefethen 2004)
                xj = grid.reshape((Nx, 1))
                xi = grid.reshape((1, Nx))
                dist = (xj - xi) / C
                np.fill_diagonal(dist, 1)  # Ignore product when i==j
                self.weights.append((1.0 / np.prod(dist, axis=1)).astype(np.float32))  # (Nx,)

            # Cartesian product of univariate grids
            if not self.reduced:
                self.xi = np.empty((np.prod(grid_sizes), self.xdim()), dtype=np.float32)
                for i, ele in enumerate(itertools.product(*self.x_grids)):
                    self.xi[i, :] = ele

    def refine(self, beta, auto=True, x_refine: np.ndarray = None):
        """Return a new interpolator with one dimension refined by one level, specified by `beta`.

        !!! Note
            If `self.reduced=True` or `auto=False`, then this function will return tuple indices `idx` corresponding
            to the new interpolation points `x`. The tuple indices specify one index along each input dimension.

        :param beta: the new refinement level indices
        :param auto: whether to automatically compute model at refinement points
        :param x_refine: `(Nx,)` use this array as the refined 1d grid if provided, otherwise compute via `leja_1d`
        :returns: `interp` - a `LagrangeInterpolator` with a refined grid (default), otherwise if `auto=False`,
                  returns `idx, x, interp`, where `idx` and `x` correspond to new interpolation points.
        """
        try:
            # Initialize a new interpolator with the new refinement levels
            interp = LagrangeInterpolator(beta, self.x_vars, model=self._model, model_args=self._model_args,
                                          model_kwargs=self._model_kwargs, init_grids=False, reduced=self.reduced)

            # Find the dimension and number of new points to add
            old_grid_sizes = self.get_grid_sizes(self.beta)
            new_grid_sizes = interp.get_grid_sizes(beta)
            dim_refine = 0
            num_new_pts = 0
            for idx, grid_size in enumerate(new_grid_sizes):
                if grid_size != old_grid_sizes[idx]:
                    dim_refine = idx
                    num_new_pts = grid_size - old_grid_sizes[idx]
                    break

            # Add points to leja grid in this dimension
            interp.x_grids = copy.deepcopy(self.x_grids)
            xi = copy.deepcopy(x_refine) if x_refine is not None else self.leja_1d(num_new_pts,
                                                                                   interp.x_vars[dim_refine].bounds(),
                                                                                   z_pts=interp.x_grids[dim_refine],
                                                                                   wt_fcn=interp.x_vars[dim_refine].pdf)
            interp.x_grids[dim_refine] = xi.astype(np.float32)

            # Update barycentric weights in this dimension
            interp.weights = copy.deepcopy(self.weights)
            Nx_old = old_grid_sizes[dim_refine]
            Nx_new = new_grid_sizes[dim_refine]
            old_wts = copy.deepcopy(self.weights[dim_refine])
            new_wts = np.zeros(Nx_new, dtype=np.float32)
            new_wts[:Nx_old] = old_wts
            bds = interp.x_vars[dim_refine].bounds()
            C = (bds[1] - bds[0]) / 4.0  # Interval capacity
            xi = interp.x_grids[dim_refine]
            for j in range(Nx_old, Nx_new):
                new_wts[:j] *= (C / (xi[:j] - xi[j]))
                new_wts[j] = np.prod(C / (xi[j] - xi[:j]))
            interp.weights[dim_refine] = new_wts

            # Copy yi over at existing interpolation points
            x_new = np.zeros((0, interp.xdim()), dtype=np.float32)
            x_new_idx = []
            tol = 1e-12     # Tolerance for floating point comparison
            j = 0           # Use this idx for iterating over existing yi
            if not self.reduced:
                interp.xi = np.zeros((np.prod(new_grid_sizes), self.xdim()), dtype=np.float32)
                interp.yi = np.zeros((np.prod(new_grid_sizes), self.ydim()), dtype=np.float32)
                if self.save_enabled():
                    interp.output_files = [None] * np.prod(new_grid_sizes)

            old_indices = [np.arange(old_grid_sizes[n]) for n in range(self.xdim())]
            old_indices = list(itertools.product(*old_indices))
            new_indices = [np.arange(new_grid_sizes[n]) for n in range(self.xdim())]
            new_indices = list(itertools.product(*new_indices))
            for i in range(len(new_indices)):
                # Get the new grid coordinate/index and physical x location/point
                new_x_idx = new_indices[i]
                new_x_pt = np.array([float(interp.x_grids[n][new_x_idx[n]]) for n in range(self.xdim())],
                                    dtype=np.float32)

                if not self.reduced:
                    # Store the old xi/yi and return new x points
                    interp.xi[i, :] = new_x_pt
                    if j < len(old_indices) and np.all(np.abs(np.array(old_indices[j]) -
                                                              np.array(new_indices[i])) < tol):
                        # If we already have this interpolation point
                        interp.yi[i, :] = self.yi[j, :]
                        if self.save_enabled():
                            interp.output_files[i] = self.output_files[j]
                        j += 1
                    else:
                        # Otherwise, save new interpolation point and its index
                        x_new = np.concatenate((x_new, new_x_pt.reshape((1, self.xdim()))))
                        x_new_idx.append(i)
                else:
                    # Just find the new x indices and return those for the reduced case
                    if j < len(old_indices) and np.all(np.abs(np.array(old_indices[j]) -
                                                              np.array(new_indices[i])) < tol):
                        j += 1
                    else:
                        x_new = np.concatenate((x_new, new_x_pt.reshape((1, self.xdim()))))
                        x_new_idx.append(new_x_idx)     # Add a tuple() multi-index if not saving xi/yi

            # Evaluate the model at new interpolation points
            interp.model_cost = self.model_cost
            if self._model is None:
                self.logger.warning(f'No model available to evaluate new interpolation points, returning the points '
                                    f'to you instead...')
                return x_new_idx, x_new, interp
            elif not auto or self.reduced:
                return x_new_idx, x_new, interp
            else:
                interp.set_yi(x_new=(x_new_idx, x_new))
                return interp

        except Exception as e:
            import traceback
            tb_str = str(traceback.format_exception(e))
            self.logger.error(tb_str)
            raise Exception(f'Original exception in refine(): {tb_str}')

    def __call__(self, x: np.ndarray | float, xi: np.ndarray = None, yi: np.ndarray = None) -> np.ndarray:
        """Evaluate the barycentric interpolation at points `x`.

        :param x: `(..., xdim)`, the points to be interpolated, must be within domain of `self.xi` for accuracy
        :param xi: `(Ni, xdim)` optional, interpolation grid points to use (e.g. if `self.reduced=True`)
        :param yi: `(Ni, ydim)` optional, function values at xi to use (e.g. if `self.reduced=True`)
        :returns y: `(..., ydim)`, the interpolated function values
        """
        x = np.atleast_1d(x)
        if yi is None:
            yi = self.yi.copy()
        if xi is None:
            xi = self.xi.copy()
        ydim = yi.shape[-1]
        nan_idx = np.any(np.isnan(yi), axis=-1)
        if np.any(nan_idx):
            # Use a simple linear regression fit to impute missing values (may have resulted from bad model outputs)
            imputer = Pipeline([('scaler', MaxAbsScaler()), ('model', Ridge(alpha=1))])
            imputer.fit(xi[~nan_idx, :], yi[~nan_idx, :])
            yi[nan_idx, :] = imputer.predict(xi[nan_idx, :])

        # Loop over multi-indices and compute tensor-product lagrange polynomials
        grid_sizes = self.get_grid_sizes(self.beta)
        y = np.zeros(x.shape[:-1] + (ydim,))    # (..., ydim)
        indices = [np.arange(grid_sizes[n]) for n in range(self.xdim())]
        for i, j in enumerate(itertools.product(*indices)):
            L_j = np.empty(x.shape)             # (..., xdim)

            # Compute univariate Lagrange polynomials in each dimension
            for n in range(self.xdim()):
                x_n = x[..., n, np.newaxis]     # (..., 1)
                x_j = self.x_grids[n]           # (Nx,)
                w_j = self.weights[n]           # (Nx,)

                # Compute the jth Lagrange basis polynomial L_j(x_n) for this x dimension (in barycentric form)
                c = x_n - x_j
                div_zero_idx = np.abs(c) <= 1e-4 * np.abs(x_j) + 1e-8   # Track where x is at an interpolation pnt x_j
                c[div_zero_idx] = 1                             # Temporarily set to 1 to avoid divide by zero error
                c = w_j / c
                L_j[..., n] = c[..., j[n]] / np.sum(c, axis=-1)  # (...) same size as original x

                # Set L_j(x==x_j)=1 for the current j and set L_j(x==x_j)=0 for x_j = x_i, i != j
                L_j[div_zero_idx[..., j[n]], n] = 1
                L_j[np.any(div_zero_idx[..., [idx for idx in range(grid_sizes[n]) if idx != j[n]]], axis=-1), n] = 0

            # Add multivariate basis polynomial contribution to interpolation output
            L_j = np.prod(L_j, axis=-1, keepdims=True)      # (..., 1)
            y += L_j * yi[i, :]

        return y

    @staticmethod
    def get_grid_sizes(beta: tuple, k: int = 2) -> list[int]:
        """Compute number of grid points in each input dimension.

        :param beta: refinement level indices
        :param k: level-to-grid-size multiplier (probably just always `k=2`)
        :returns: list of grid sizes in each dimension
        """
        return [k*beta[i] + 1 for i in range(len(beta))]

    @staticmethod
    def leja_1d(N: int, z_bds: tuple, z_pts: np.ndarray = None, wt_fcn: callable = None) -> np.ndarray:
        """Find the next `N` points in the Leja sequence of `z_pts`.

        :param N: number of new points to add to the sequence
        :param z_bds: bounds on the 1d domain
        :param z_pts: current univariate Leja sequence `(Nz,)`, start at middle of `z_bds` if `None`
        :param wt_fcn: weighting function, uses a constant weight if `None`, callable as `wt_fcn(z)`
        :returns: the Leja sequence `z_pts` augmented by `N` new points
        """
        # if wt_fcn is None:
        wt_fcn = lambda z: 1  # UPDATE: ignore RV weighting, unbounded pdfs like Gaussian cause problems
        if z_pts is None:
            z_pts = (z_bds[1] + z_bds[0]) / 2
            N = N - 1
        z_pts = np.atleast_1d(z_pts).astype(np.float32)

        # Construct Leja sequence by maximizing the Leja objective sequentially
        for i in range(N):
            obj_fun = lambda z: -wt_fcn(np.array(z).astype(np.float32)) * np.prod(np.abs(z - z_pts))
            res = direct(obj_fun, [z_bds])  # Use global DIRECT optimization over 1d domain
            z_star = res.x
            z_pts = np.concatenate((z_pts, z_star))

        return z_pts



import time
import warnings
from pathlib import Path
import sys

import numpy as np
from scipy.linalg import lapack
from scipy.optimize import fsolve
from scipy.stats import gaussian_kde
import matplotlib.pyplot as plt
import pytest

from amisc.system import SystemSurrogate, ComponentSpec
from amisc.rv import UniformRV
from amisc.utils import ax_default
from amisc.examples.models import fire_sat_system


# TODO: Include a swap and insert component test

@pytest.mark.skipif(not sys.platform.startswith('linux'), reason='not sure why')
def test_fire_sat(plots=True):
    """Test the fire satellite coupled system from Chaudhuri (2018)"""
    N = 100
    surr = fire_sat_system(save_dir=Path('.'))
    xt = surr.sample_inputs(N, use_pdf=True)
    yt = surr(xt, use_model='best')
    use_idx = ~np.any(np.isnan(yt), axis=-1)
    xt = xt[use_idx, :]
    yt = yt[use_idx, :]
    test_set = {'xt': xt, 'yt': yt}
    n_jobs = -1 if sys.platform.startswith('linux') else 1  # parallel on VM runners on Github get stuck for non-linux
    surr.fit(max_iter=15, max_tol=1e-2, max_runtime=1/12, test_set=test_set, n_jobs=n_jobs,
             num_refine=1000)

    ysurr = surr.predict(xt)
    l2_error = np.sqrt(np.nanmean((ysurr-yt)**2, axis=0)) / np.sqrt(np.nanmean(yt**2, axis=0))
    assert np.nanmax(l2_error) < 0.1

    if plots:
        # Plot error histograms
        fig, ax = plt.subplots(1, 3)
        ax[0].hist(yt[:, 0], color='red', bins=20, edgecolor='black', density=True, linewidth=1.2, label='Truth')
        ax[0].hist(ysurr[:, 0], color='blue', bins=20, edgecolor='black', density=True, linewidth=1.2, alpha=0.4, label='Surrogate')
        ax[1].hist(yt[:, 7], color='red', bins=20, edgecolor='black', density=True, linewidth=1.2, label='Truth')
        ax[1].hist(ysurr[:, 7], color='blue', bins=20, edgecolor='black', density=True, linewidth=1.2, alpha=0.4, label='Surrogate')
        ax[2].hist(yt[:, 8], color='red', bins=20, edgecolor='black', density=True, linewidth=1.2, label='Truth')
        ax[2].hist(ysurr[:, 8], color='blue', bins=20, edgecolor='black', density=True, linewidth=1.2, alpha=0.4, label='Surrogate')
        ax_default(ax[0], 'Satellite velocity ($m/s$)', '', legend=True)
        ax_default(ax[1], 'Solar panel area ($m^2$)', '', legend=True)
        ax_default(ax[2], 'Attitude control power ($W$)', '', legend=True)
        fig.set_size_inches(9, 3)
        fig.tight_layout()

        # Plot 1d slices
        slice_idx = ['H', 'Po', 'Cd']
        qoi_idx = ['Vsat', 'Asa', 'Pat']
        try:
            fig, ax = surr.plot_slice(slice_idx, qoi_idx, show_model=['best', 'worst'], model_dir=surr.root_dir,
                                      random_walk=True, N=10)
        except np.linalg.LinAlgError:
            print(f'Its alright. Sometimes the random walks are wacky and FPI wont converge.')

        # Plot allocation bar charts
        fig, ax = surr.plot_allocation()
        plt.show()


def test_system_refine(plots=False):
    """Test iterative refinement for Figure 5 in Jakeman 2022"""
    def coupled_system():
        def f1(x):
            return {'y': x * np.sin(np.pi * x)}
        def f2(x):
            return {'y': 1 / (1 + 25 * x ** 2)}
        return f1, f2

    f1, f2 = coupled_system()
    exo_vars = [UniformRV(0, 1)]
    coupling_vars = [UniformRV(0, 1), UniformRV(0, 1)]
    comp1 = ComponentSpec(f1, name='Model1', exo_in=0, coupling_out=0)
    comp2 = ComponentSpec(f2, name='Model2', coupling_in=0, coupling_out=1)
    surr = SystemSurrogate([comp1, comp2], exo_vars, coupling_vars, init_surr=True)

    Niter = 4
    x = np.linspace(0, 1, 100).reshape((100, 1))
    y1 = f1(x)['y']
    y2 = f2(x)['y']
    y3 = f2(y1)['y']
    fig, ax = plt.subplots(Niter, 3, sharex='col', sharey='row')
    for i in range(Niter):
        # Plot actual function values
        ax[i, 0].plot(x, y1, '-r', label='$f_1(x)$')
        ax[i, 1].plot(x, y2, '-r', label='$f_2(y)$')
        ax[i, 2].plot(x, y3, '-r', label='$f(x)$')

        # Plot first component surrogates
        comp = surr.get_component('Model1')
        ax[i, 0].plot(x, comp(x, training=True), '--k', label='$f_1$ current')
        beta_max = 0
        for alpha, beta in comp.index_set:
            if beta[0] > beta_max:
                beta_max = beta[0]
        interp = comp.get_sub_surrogate((), (beta_max,), include_grid=True)
        ax[i, 0].plot(interp.xi, interp.yi, 'ok', markersize=8, label='')
        for alpha, beta in comp.iterate_candidates():
            comp.update_misc_coeffs()
            yJ1 = surr(x, training=True)
            ax[i, 0].plot(x, comp(x, training=True), ':b', label='$f_1$ candidate')
            interp = comp.get_sub_surrogate(alpha, beta, include_grid=True)
            ax[i, 0].plot(interp.xi, interp.yi, 'xb', markersize=8, label='')
        comp.update_misc_coeffs()

        # Plot second component surrogates
        comp = surr.get_component('Model2')
        ax[i, 1].plot(x, comp(x, training=True), '--k', label='$f_2$ current')
        beta_max = 0
        for alpha, beta in comp.index_set:
            if beta[0] > beta_max:
                beta_max = beta[0]
        interp = comp.get_sub_surrogate((), (beta_max,), include_grid=True)
        ax[i, 1].plot(interp.xi, interp.yi, 'ok', markersize=8, label='')
        for alpha, beta in comp.iterate_candidates():
            comp.update_misc_coeffs()
            yJ2 = surr(x, training=True)
            ax[i, 1].plot(x, comp(x, training=True), '-.g', label='$f_2$ candidate')
            interp = comp.get_sub_surrogate(alpha, beta, include_grid=True)
            ax[i, 1].plot(interp.xi, interp.yi, 'xg', markersize=8, label='')
        comp.update_misc_coeffs()

        # Plot integrated surrogates
        ysurr = surr(x, training=True)
        ax[i, 2].plot(x, ysurr[:, 1:2], '--k', label='$f_J$')
        ax[i, 2].plot(x, yJ1[:, 1:2], ':b', label='$f_{J_1}$')
        ax[i, 2].plot(x, yJ2[:, 1:2], '-.g', label='$f_{J_2}$')
        ax_default(ax[i, 0], '$x$', '$f_1(x)$', legend=True)
        ax_default(ax[i, 1], '$y$', '$f_2(y)$', legend=True)
        ax_default(ax[i, 2], '$x$', '$f_2(f_1(x))$', legend=True)

        # Refine the system
        surr.refine(qoi_ind=None, num_refine=100)

    ysurr = surr.predict(x)
    ytrue = surr.predict(x, use_model='best')
    l2_error = np.sqrt(np.mean((ysurr-ytrue)**2, axis=0)) / np.sqrt(np.mean(ytrue**2, axis=0))
    assert np.max(l2_error) < 0.1

    if plots:
        fig.set_size_inches(3.5*3, 3.5*Niter)
        fig.tight_layout()
        plt.show()


def test_feedforward(plots=False):
    """Test MD system in Figure 5 in Jakeman 2022"""
    def coupled_system():
        def f1(x):
            return {'y': x * np.sin(np.pi * x)}
        def f2(x):
            return {'y': 1 / (1 + 25*x**2)}
        def f(x):
            return f2(f1(x)['y'])['y']
        return f1, f2, f

    f1, f2, f = coupled_system()
    x_in = UniformRV(0, 1)
    y1, y2 = UniformRV(0, 1), UniformRV(0, 1)
    comp1 = ComponentSpec(f1, exo_in=x_in, coupling_out=y1)
    comp2 = ComponentSpec(f2, coupling_in=y1, coupling_out=y2)
    surr = SystemSurrogate([comp1, comp2], x_in, [y1, y2], init_surr=False)

    x = np.linspace(0, 1, 100).reshape((100, 1))
    y1 = f1(x)['y']
    y2 = f(x)
    y_surr = surr(x, use_model='best')
    l2_error = np.sqrt(np.mean((y_surr[:, 1:2] - y2)**2)) / np.sqrt(np.mean(y2**2))
    assert l2_error < 1e-15

    if plots:
        fig, ax = plt.subplots(1, 2)
        ax[0].plot(x, y1, '-r', label='$f_1(x)$')
        ax[0].plot(np.squeeze(x), y_surr[:, 0], '--k', label='Surrogate')
        ax_default(ax[0], '$x$', '$f(x)$', legend=True)
        ax[1].plot(x, y2, '-r', label='$f(x)$')
        ax[1].plot(np.squeeze(x), y_surr[:, 1], '--k', label='Surrogate')
        ax_default(ax[1], '$x$', '$f(x)$', legend=True)
        fig.set_size_inches(6, 3)
        fig.tight_layout()
        plt.show()


def test_system_surrogate(plots=False):
    """Test the MD system in figure 6 in Jakeman 2022"""
    def coupled_system(D1, D2, Q1, Q2):
        def f1(x, /, alpha=(0,)):
            eps = 10 ** (-float(alpha[0]))
            q = np.arange(1, Q1+1).reshape((1,)*len(x.shape[:-1]) + (Q1,))
            return {'y': (x[..., 0, np.newaxis] ** (q-1)) * np.sin(np.sum(x, axis=-1, keepdims=True) + eps)}

        def f2(x, /, alpha=(0,)):
            eps = 10 ** (-float(alpha[0]))
            q = np.arange(1, Q2+1).reshape((1,)*len(x.shape) + (Q2,))
            prod1 = np.prod(x[..., D2:, np.newaxis] ** (q) - eps, axis=-2)  # (..., Q2)
            prod2 = np.prod(x[..., :D2], axis=-1, keepdims=True)            # (..., 1)
            return {'y': prod1 * prod2}

        def f3(x, /, alpha=(0,), D3=D1):
            eps = 10 ** (-float(alpha[0]))
            prod1 = np.exp(-np.sum((x[..., D3:] - eps) ** 2, axis=-1))  # (...,)
            prod2 = 1 + (25/16)*np.sum(x[..., :D3], axis=-1) ** 2       # (...,)
            return {'y': np.expand_dims(prod1 / prod2, axis=-1)}        # (..., 1)

        def f(x):
            # Ground truth (arbitrary high alpha)
            alpha = (15,)
            x1 = x[..., :D1]
            y1 = f1(x1, alpha)['y']
            x2 = np.concatenate((x[..., D1:], y1), axis=-1)
            y2 = f2(x2, alpha)['y']
            x3 = np.concatenate((x1, y2), axis=-1)
            y3 = f3(x3, alpha)['y']
            return np.concatenate((y1, y2, y3), axis=-1)

        return f1, f2, f3, f

    # Hook up the 'wiring' for this example feedforward system
    D1 = 1
    D2 = D1
    Q1 = 1
    Q2 = Q1
    alpha = (15,)
    f1, f2, f3, f = coupled_system(D1, D2, Q1, Q2)
    comp1 = ComponentSpec(f1, name='Cathode', truth_alpha=alpha, exo_in=list(np.arange(0, D1)),
                          coupling_out=list(np.arange(0, Q1)), max_alpha=5, max_beta=(3,)*D1)
    comp2 = ComponentSpec(f2, name='Thruster', truth_alpha=alpha, exo_in=list(np.arange(D1, D1+D2)), max_alpha=5,
                          max_beta=(3,)*(D2+Q1), coupling_in={'Cathode': list(np.arange(0, Q1))},
                          coupling_out=list(np.arange(Q1, Q1+Q2)))
    comp3 = ComponentSpec(f3, name='Plume', truth_alpha=alpha, exo_in=list(np.arange(0, D1)), max_alpha=5,
                          coupling_in={'Thruster': list(np.arange(0, Q2))}, coupling_out=Q1+Q2, max_beta=(3,)*(D1+Q2))
    exo_vars = [UniformRV(0, 1) for i in range(D1+D2)]
    coupling_vars = [UniformRV(0, 1) for i in range(Q1+Q2+1)]
    surr = SystemSurrogate([comp1, comp2, comp3], exo_vars, coupling_vars, init_surr=False)

    # Test example
    N = 5000
    x = np.random.rand(N, D1+D2)
    y = f(x)
    y_surr = surr(x, use_model='best')
    l2_error = np.sqrt(np.mean((y_surr - y)**2, axis=0)) / np.sqrt(np.mean(y**2, axis=0))
    assert np.max(l2_error) < 1e-15

    # Show coupling variable pdfs
    if plots:
        fig, ax = plt.subplots()
        ls = ['-r', '--k', ':b']
        pts = np.linspace(0, 1, 100)
        for i in range(3):
            label_str = f'$\\rho(y_{{{i+1}}})$'
            kernel = gaussian_kde(y_surr[:, i])
            # ax[i].hist(y_surr[:, i], density=True, bins=20, color='r', edgecolor='black', linewidth=1.2)
            ax.plot(pts, kernel(pts), ls[i], label=label_str)
        ax_default(ax, r'$y$', 'PDF', legend=True)
        # fig.set_size_inches(9, 3)
        fig.tight_layout()
        plt.show()


def test_fpi():
    """Test fixed point iteration implementation against scipy fsolve."""
    f1 = lambda x: {'y': -x[..., 0:1]**3 + 2 * x[..., 1:2]**2}
    f2 = lambda x: {'y': 3*x[..., 0:1]**2 + 4 * x[..., 1:2]**(-2)}
    comp1 = ComponentSpec(f1, exo_in=[0], coupling_in={'m2': [0]}, coupling_out=[0], max_beta=(5,), name='m1')
    comp2 = ComponentSpec(f2, exo_in=[0], coupling_in={'m1': [0]}, coupling_out=[1], max_beta=(5,), name='m2')
    exo_vars = UniformRV(0, 4)
    coupling_vars = [UniformRV(1, 10), UniformRV(1, 10)]
    surr = SystemSurrogate([comp1, comp2], exo_vars, coupling_vars, init_surr=False)

    # Test on random x against scipy.fsolve
    N = 100
    tol = 1e-12
    x0 = np.array([5.5, 5.5])
    exo = surr.sample_inputs(N)
    y_surr = surr.predict(exo, use_model='best', anderson_mem=10, max_fpi_iter=200, fpi_tol=tol)  # (N, 2)
    nan_idx = list(np.any(np.isnan(y_surr), axis=-1).nonzero()[0])
    y_true = np.zeros((N, 2))
    bad_idx = []
    warnings.simplefilter('error')
    for i in range(N):
        def fun(x):
            y1 = x[0]
            y2 = x[1]
            res1 = -exo[i, 0]**3 + 2*y2**2 - y1
            res2 = 3*exo[i, 0]**2 + 4*y1**(-2) - y2
            return [res1, res2]

        try:
            y_true[i, :] = fsolve(fun, x0, xtol=tol)
        except Exception as e:
            bad_idx.append(i)

    y_surr = np.delete(y_surr, nan_idx + bad_idx, axis=0)
    y_true = np.delete(y_true, nan_idx + bad_idx, axis=0)
    l2_error = np.sqrt(np.mean((y_surr - y_true)**2, axis=0)) / np.sqrt(np.mean(y_true**2, axis=0))
    assert np.max(l2_error) < tol


def test_lls():
    """Test constrained linear least squares routine against scipy lapack."""
    X = 100
    Y = 100
    M = 10
    N = 10
    P = 1
    tol = 1e-8

    A = np.random.rand(X, Y, M, N)
    b = np.random.rand(X, Y, M, 1)
    C = np.random.rand(X, Y, P, N)
    d = np.random.rand(X, Y, P, 1)

    # custom solver
    t1 = time.time()
    alpha = np.squeeze(SystemSurrogate._constrained_lls(A, b, C, d), axis=-1)  # (*, N)
    t2 = time.time()

    # Built in scipy solver
    alpha2 = np.zeros((X, Y, N))
    t3 = time.time()
    for i in range(X):
        for j in range(Y):
            Ai = A[i, j, ...]
            bi = b[i, j, ...]
            Ci = C[i, j, ...]
            di = d[i, j, ...]
            ret = lapack.dgglse(Ai, Ci, bi, di)
            alpha2[i, j, :] = ret[3]
    t4 = time.time()

    # Results
    diff = alpha - alpha2
    assert np.max(np.abs(diff)) < tol
    print(f'Custom CLLS time: {t2-t1} s. Scipy time: {t4-t3} s.')



"""`rv.py`

Provides small classes for random variables.

Includes
--------
- `BaseRV`: Abstract wrapper class of a random variable
- `UniformRV`: a uniformly-distributed random variable
- `NormalRV`: a normally-distributed random variable
- `ScalarRV`: a stand-in class for a variable with no uncertainty or pdf
- `LogUniformRV`: base 10 log-uniform
- `LogNormalRV`: base 10 log-normal
"""
from abc import ABC, abstractmethod
import random
import string

import numpy as np


class BaseRV(ABC):
    """Small wrapper class similar to `scipy.stats` random variables (RVs).

    :ivar id: an identifier for the variable
    :ivar bds: the explicit domain bounds of the variable (limits of where you expect to use it)
    :ivar nominal: a typical value for this variable (within `bds`)
    :ivar tex: latex format for the random variable, i.e. r"$x_i$"
    :ivar description: a lengthier description of the variable
    :ivar units: assumed units for the variable (if applicable)
    :ivar param_type: an additional descriptor for how this rv is used, e.g. calibration, operating, design, etc.

    :vartype id: str
    :vartype bds: tuple[float, float]
    :vartype nominal: float
    :vartype tex: str
    :vartype description: str
    :vartype units: str
    :vartype param_type: str
    """

    def __init__(self, id='', *, tex='', description='Random variable', units='-',
                 param_type='calibration', nominal=1, domain=(0, 1)):
        """Child classes must implement `sample` and `pdf` methods."""
        self.bds = tuple(domain)
        self.nominal = nominal
        self.id = id if id != '' else 'X_' + ''.join(random.choices(string.ascii_uppercase + string.digits, k=4))
        self.is_custom_id = id != ''  # Whether a custom id was assigned to this variable
        self.tex = tex
        self.description = description
        self.units = units
        self.param_type = param_type

    def __repr__(self):
        return r'{}'.format(f"{self.id} - {self.description} ({self.units})")

    def __str__(self):
        return self.id

    def __eq__(self, other):
        """Consider two RVs equal if they share the same string id.

        Also returns true when checking if this RV is equal to a string id by itself.
        """
        if isinstance(other, BaseRV):
            return self.id == other.id
        elif isinstance(other, str):
            return self.id == other
        else:
            return NotImplemented

    def __hash__(self):
        return hash(self.id)

    def to_tex(self, units=False, symbol=True):
        """Return a raw string that is well-formatted for plotting (with tex).

        :param units: whether to include the units in the string
        :param symbol: just latex symbol if true, otherwise the full description
        """
        s = self.tex if symbol else self.description
        if s == '':
            s = str(self)
        return r'{} [{}]'.format(s, self.units) if units else r'{}'.format(s)

    def bounds(self):
        """Return a tuple of the defined domain of this RV."""
        return self.bds

    def update_bounds(self, lb, ub):
        """Update the defined domain of this RV to `(lb, ub)`."""
        self.bds = (lb, ub)

    def sample_domain(self, shape: tuple | int) -> np.ndarray:
        """Return an array of the given `shape` for random samples over the domain of this RV.

        :param shape: the shape of samples to return
        :returns samples: random samples over the domain of the random variable
        """
        if isinstance(shape, int):
            shape = (shape, )
        return np.random.rand(*shape) * (self.bds[1] - self.bds[0]) + self.bds[0]

    @abstractmethod
    def pdf(self, x: np.ndarray) -> np.ndarray:
        """Compute the PDF of the RV at the given `x` locations.

        :param x: locations to compute the PDF at
        :returns f: the PDF evaluations at `x`
        """
        pass

    @abstractmethod
    def sample(self, shape: tuple | int, nominal: float = None) -> np.ndarray:
        """Draw samples from the PDF.

        :param shape: the shape of the returned samples
        :param nominal: a nominal value to use if applicable (i.e. a center for relative Uniform or Normal)
        :returns: samples from the PDF of this random variable
        """
        pass


class ScalarRV(BaseRV):
    """A stand-in variable with no uncertainty/pdf, just scalars."""

    def pdf(self, x):
        return np.ones(x.shape)

    def sample(self, shape, nominal=None):
        if isinstance(shape, int):
            shape = (shape, )
        if nominal is not None:
            return np.ones(shape)*nominal
        else:
            return self.sample_domain(shape)


class UniformRV(BaseRV):
    """A uniformly-distributed random variable.

    Can be uniformly distributed in one of three ways: between global bounds, relative within a percent, or relative
    within a set absolute tolerance.

    :ivar type: specifies the type of uniform distribution, either 'bds', 'pct', or 'tol' as described above
    :ivar value: the absolute tolerance or percent uncertainty if type is 'tol' or 'pct'

    :vartype type: str
    :vartype value: float
    """

    def __init__(self, arg1: float, arg2: float | str, id='', **kwargs):
        """Construct a uniformly-distributed random variable.

        :param arg1: lower bound if specifying U(lb, ub), otherwise a tol or pct if specifying U(+/- tol/pct)
        :param arg2: upper bound if specifying U(lb, ub), otherwise a str of either 'tol' or 'pct'
        """
        domain = kwargs.get('domain', None)
        if isinstance(arg2, str):
            self.value = arg1
            self.type = arg2
        else:
            self.value = None
            self.type = 'bds'
        if self.type == 'bds':
            domain = (arg1, arg2) if domain is None else tuple(domain)     # This means domain overrides (arg1, arg2)
        else:
            domain = (0, 1) if domain is None else tuple(domain)
        kwargs['domain'] = domain
        super().__init__(id, **kwargs)

        # Set default nominal value as middle of the domain if not specified
        if kwargs.get('nominal', None) is None:
            self.nominal = (self.bds[1] + self.bds[0]) / 2

    def __str__(self):
        return self.id if self.is_custom_id else f'U({self.bds[0]}, {self.bds[1]})'

    def get_uniform_bounds(self, nominal: float = None) -> tuple[float, float]:
        """Return the correct set of bounds based on type of uniform distribution.

        :param nominal: the center value for relative uniform distributions
        :returns: the uniform bounds
        """
        match self.type:
            case 'bds':
                return self.bds
            case 'pct':
                if nominal is None:
                    return self.bds
                return nominal * (1 - self.value), nominal * (1 + self.value)
            case 'tol':
                if nominal is None:
                    return self.bds
                return nominal - self.value, nominal + self.value
            case other:
                raise NotImplementedError(f'self.type = {self.type} not known. Choose from ["pct, "tol", "bds"]')

    def pdf(self, x: np.ndarray, nominal: float = None) -> np.ndarray:
        """Compute the pdf for a uniform distribution.

        :param x: locations to compute the pdf at
        :param nominal: center location for relative uniform rvs
        :returns: the evaluated PDF at `x`
        """
        bds = self.get_uniform_bounds(nominal)
        den = bds[1] - bds[0]
        den = 1 if np.isclose(den, 0) else den
        y = np.broadcast_to(1 / den, x.shape).copy()
        y[np.where(x > bds[1])] = 0
        y[np.where(x < bds[0])] = 0
        return y

    def sample(self, shape, nominal=None):
        if isinstance(shape, int):
            shape = (shape, )
        bds = self.get_uniform_bounds(nominal)
        return np.random.rand(*shape) * (bds[1] - bds[0]) + bds[0]


class LogUniformRV(BaseRV):
    """A base 10 log-uniform distributed random variable, only supports absolute bounds."""

    def __init__(self, log10_a: float, log10_b: float, id='', **kwargs):
        """Construct the log-uniform random variable.

        :param log10_a: the lower bound in log10 space
        :param log10_b: the upper bound in log10 space
        """
        super().__init__(id, **kwargs)
        self.bds = (10**log10_a, 10**log10_b)

    def __str__(self):
        return self.id if self.is_custom_id else f'LU({np.log10(self.bds[0]): .2f}, {np.log10(self.bds[1]): .2f})'

    def pdf(self, x):
        return np.log10(np.e) / (x * (np.log10(self.bds[1]) - np.log10(self.bds[0])))

    def sample(self, shape, nominal=None):
        if isinstance(shape, int):
            shape = (shape, )
        lb = np.log10(self.bds[0])
        ub = np.log10(self.bds[1])
        return 10 ** (np.random.rand(*shape) * (ub - lb) + lb)


class LogNormalRV(BaseRV):
    """A base 10 log-normal distributed random variable.

    :ivar mu: the center of the log-normal distribution
    :ivar std: the standard deviation of the log-normal distribution

    :vartype mu: float
    :vartype std: float
    """

    def __init__(self, mu: float, std: float, id='', **kwargs):
        """Construct the RV with the mean and std of the underlying distribution,
        i.e. $\\log_{10}(x) \\sim N(\\mu, \\sigma)$.

        :param mu: the center of the log-normal distribution
        :param std: the standard deviation of the log-normal distribution
        """
        domain = kwargs.get('domain', None)
        if domain is None:
            domain = (10 ** (mu - 3*std), 10 ** (mu + 3*std))   # Use a default domain of +- 3std
        kwargs['domain'] = domain
        super().__init__(id, **kwargs)
        self.std = std
        self.mu = mu

    def recenter(self, mu: float, std: float = None):
        """Move the center of the distribution to `mu` with standard deviation `std` (optional)

        :param mu: the new center of the distribution
        :param std: (optional) new standard deviation
        """
        self.mu = mu
        if std is not None:
            self.std = std

    def __str__(self):
        return self.id if self.is_custom_id else f'LN_10({self.mu}, {self.std})'

    def pdf(self, x):
        return (np.log10(np.e) / (x * self.std * np.sqrt(2 * np.pi))) * \
               np.exp(-0.5 * ((np.log10(x) - self.mu) / self.std) ** 2)

    def sample(self, shape, nominal=None):
        if isinstance(shape, int):
            shape = (shape, )
        scale = np.log10(np.e)
        center = self.mu if nominal is None else nominal
        return np.random.lognormal(mean=(1 / scale) * center, sigma=(1 / scale) * self.std, size=shape)
        # return 10 ** (np.random.randn(*size)*self.std + center)  # Alternatively


class NormalRV(BaseRV):
    """A normally-distributed random variable.

    :ivar mu: float, the mean of the normal distribution
    :ivar std: float, the standard deviation of the normal distribution

    :vartype mu: float
    :vartype std: float
    """

    def __init__(self, mu, std, id='', **kwargs):
        domain = kwargs.get('domain', None)
        if domain is None:
            domain = (mu - 2.5*std, mu + 2.5*std)   # Use a default domain of +- 2.5std
        kwargs['domain'] = domain
        super().__init__(id, **kwargs)
        self.mu = mu
        self.std = std

        # Set default nominal value as the provided mean
        if kwargs.get('nominal', None) is None:
            self.nominal = mu

    def recenter(self, mu: float, std: float =None):
        """Move the center of the distribution to `mu` with standard deviation `std` (optional)

        :param mu: the new center of the distribution
        :param std: (optional) new standard deviation
        """
        self.mu = mu
        if std is not None:
            self.std = std

    def __str__(self):
        return self.id if self.is_custom_id else f'N({self.mu}, {self.std})'

    def pdf(self, x):
        return (1 / (np.sqrt(2 * np.pi) * self.std)) * np.exp(-0.5 * ((x - self.mu) / self.std) ** 2)

    def sample(self, shape, nominal=None):
        if isinstance(shape, int):
            shape = (shape, )
        center = self.mu if nominal is None else nominal
        return np.random.randn(*shape) * self.std + center



import numpy as np
import matplotlib.pyplot as plt

from amisc.utils import approx_hess, batch_normal_sample, ax_default, get_logger, ndscatter


def test_hessian():
    """Test the Hessian of $f(x, y) = 2x^2 + 3xy^3$"""
    shape = (11, 12)
    dim = 2
    theta0 = np.random.rand(*shape, dim)
    fun = lambda theta: 2 * theta[..., 0:1]**2 + 3 * theta[..., 0:1] * theta[..., 1:2]**3

    H_tilde = approx_hess(fun, theta0)
    H_exact = np.empty(shape + (dim, dim))
    H_exact[..., 0, 0] = 4
    H_exact[..., 0, 1] = 9 * theta0[..., 1]**2
    H_exact[..., 1, 0] = 9 * theta0[..., 1]**2
    H_exact[..., 1, 1] = 18 * theta0[..., 0] * theta0[..., 1]

    assert np.allclose(H_tilde, H_exact, rtol=1e-3, atol=1e-3)


def test_sample():
    """Test 1d and 2d batch normal sampling"""
    dim = 2
    shape = (4, 5)
    N = 100000
    mean = np.random.rand(*shape, dim)
    cov = np.eye(dim) * 0.01
    samples = batch_normal_sample(mean, cov, N)
    assert samples.shape == (N, *shape, dim)
    assert np.allclose(np.mean(samples, axis=0), mean, rtol=1e-3, atol=1e-3)

    mean = np.random.rand()
    cov = 0.01
    samples = batch_normal_sample(mean, cov, N)
    assert np.isclose(mean, np.mean(samples, axis=0), rtol=1e-3, atol=1e-3)


def test_logging_and_plotting():
    """Test logging and plotting utils"""
    x = np.linspace(0, 1, 100)
    fig, ax = plt.subplots()
    ax.plot(x, x, label='Hey there:)')
    ax_default(ax, 'X label here', 'Y label here', legend=True)
    logger = get_logger('tester', stdout=True)
    logger.info('Testing logger...')

    mean = np.array([5, -2])
    cov = np.array([[2, 0.4], [0.2, 1]])
    samples = np.random.multivariate_normal(mean, cov.T @ cov, size=100)
    yt = samples[:, 0] + samples[:, 1] ** 2
    ysurr = yt + np.random.randn(*yt.shape)
    err = np.abs(ysurr - yt) / np.abs(yt)
    ndscatter(samples, labels=[r'$\alpha$', r'$\beta$', r'$\gamma$'], plot='scatter', cmap='plasma',
              cb_norm='log', z=err)



"""`system.py`

The `SystemSurrogate` is a framework for multidisciplinary models. It manages multiple single discipline component
models and the connections between them. It provides a top-level interface for constructing and evaluating surrogates.

Features
--------
- Manages multidisciplinary models in a graph data structure, supports feedforward and feedback connections
- Feedback connections are solved with a fixed-point iteration (FPI) nonlinear solver
- FPI uses Anderson acceleration and surrogate evaluations for speed-up
- Top-level interface for training and using surrogates of each component model
- Adaptive experimental design for choosing training data efficiently
- Convenient testing, plotting, and performance metrics provided to assess quality of surrogates
- Detailed logging and traceback information
- Supports parallel execution with OpenMP and MPI protocols
- Abstract and flexible interfacing with component models

!!! Info "Model specification"
    Models are callable Python wrapper functions of the form `ret = model(x, *args, **kwargs)`, where `x` is an
    `np.ndarray` of model inputs (and `*args, **kwargs` allow passing any other required configurations for your model).
    The return value is a Python dictionary of the form `ret = {'y': y, 'files': files, 'cost': cost, etc.}`.
    In the return dictionary, you specify the raw model output `y` as an `np.ndarray` at a _minimum_. Optionally, you can
    specify paths to output files and the average model cost (in seconds of cpu time), and anything else you want. Your
    `model()` function can do anything it wants in order to go from `x` &rarr; `y`. Python has the flexibility to call
    virtually any external codes, or to implement the function natively with `numpy`.

!!! Info "Component specification"
    A component adds some extra configuration around a callable `model`. These configurations are defined in a Python
    dictionary, which we give the custom type `ComponentSpec`. At a bare _minimum_, you must specify a callable
    `model` and its connections to other models within the multidisciplinary system. The limiting case is a single
    component model, for which the configuration is simply `component = ComponentSpec(model)`.
"""
import os
import time
import datetime
import functools
import copy
from datetime import timezone
from pathlib import Path
import random
import string
import pickle
from collections import UserDict
from concurrent.futures import Executor

import numpy as np
import networkx as nx
import dill
import matplotlib.pyplot as plt
from joblib import Parallel, delayed
from joblib.externals.loky import set_loky_pickler

from amisc.component import SparseGridSurrogate, ComponentSurrogate, AnalyticalSurrogate
from amisc import IndicesRV, IndexSet
from amisc.utils import ax_default, get_logger
from amisc.rv import BaseRV


class ComponentSpec(UserDict):
    """Provides a simple extension class of a Python dictionary, used to configure a component model.

    !!! Info "Specifying a list of random variables"
        The three fields: `exo_in`, `coupling_in`, and `coupling_out` fully determine how a component fits within a
        multidisciplinary system. For each, you must specify a list of variables in the same order as the model uses
        them. The model will use all exogenous inputs first, and then all coupling inputs. You can use a variable's
        global integer index into the system `exo_vars` or `coupling_vars`, or you can use the `str` id of the variable
        or the variable itself. This is summarized in the `amisc.IndicesRV` custom type.

    !!! Example
        Let's say you have a model:
        ```python
        def my_model(x, *args, **kwargs):
            print(x.shape)  # (3,), so a total of 3 inputs
            G = 6.674e-11
            m1 = x[0]           # System-level input
            m2 = x[1]           # System-level input
            r = x[2]            # Coupling input
            F = G*m1*m2 / r**2
            return {'y': F}
        ```
        Let's say this model is part of a larger system where `m1` and `m2` are specified by the system, and `r` comes
        from a different model that predicts the distance between two objects. You would set the configuration as:
        ```python
        component = ComponentSpec(my_model, exo_in=['m1', 'm2'], coupling_in=['r'], coupling_out=['F'])
        ```
    """
    Options = ['model', 'name', 'exo_in', 'coupling_in', 'coupling_out', 'truth_alpha', 'max_alpha', 'max_beta',
               'surrogate', 'model_args', 'model_kwargs', 'save_output']

    def __init__(self, model: callable, name: str = '', exo_in: IndicesRV = None,
                 coupling_in: IndicesRV | dict[str: IndicesRV] = None, coupling_out: IndicesRV = None,
                 truth_alpha: tuple | int = (), max_alpha: tuple | int = (), max_beta: tuple | int = (),
                 surrogate: str | ComponentSurrogate = 'lagrange', model_args: tuple = (), model_kwargs: dict = None,
                 save_output: bool = False):
        """Construct the configuration for this component model.

        !!! Warning
            Always specify the model at a _global_ scope, i.e. don't use `lambda` or nested functions. When saving to
            file, only a symbolic reference to the function signature will be saved, which must be globally defined
            when loading back from that save file.

        :param model: the component model, must be defined in a global scope (i.e. in a module or top-level of a script)
        :param name: the name used to identify this component model
        :param exo_in: specifies the global, system-level (i.e. exogenous/external) inputs to this model
        :param coupling_in: specifies the coupling inputs received from other models
        :param coupling_out: specifies all outputs of this model (which may couple later to downstream models)
        :param truth_alpha: the model fidelity indices to treat as a "ground truth" reference
        :param max_alpha: the maximum model fidelity indices to allow for refinement purposes
        :param max_beta: the maximum surrogate fidelity indices to allow for refinement purposes
        :param surrogate: one of ('lagrange, 'analytical'), or the `ComponentSurrogate` class to use directly
        :param model_args: optional arguments to pass to the component model
        :param model_kwargs: optional keyword arguments to pass to the component model
        :param save_output: whether this model will be saving outputs to file
        """
        d = locals()
        d2 = {key: value for key, value in d.items() if key in ComponentSpec.Options}
        super().__init__(d2)

    def __setitem__(self, key, value):
        if key in ComponentSpec.Options:
            super().__setitem__(key, value)
        else:
            raise ValueError(f'"{key}" is not applicable for a ComponentSpec. Try one of {ComponentSpec.Options}.')

    def __delitem__(self, key):
        raise TypeError("Not allowed to delete items from a ComponentSpec.")


class SystemSurrogate:
    """Multidisciplinary (MD) surrogate framework top-level class.

    !!! Note "Accessing individual components"
        The `ComponentSurrogate` objects that compose `SystemSurrogate` are internally stored in the `self.graph.nodes`
        data structure. You can access them with `get_component(comp_name)`.

    :ivar exo_vars: global list of exogenous/external inputs for the MD system
    :ivar coupling_vars: global list of coupling variables for the MD system (including all system-level outputs)
    :ivar refine_level: the total number of refinement steps that have been made
    :ivar build_metrics: contains data that summarizes surrogate training progress
    :ivar root_dir: root directory where all surrogate build products are saved to file
    :ivar log_file: log file where all logs are written to by default
    :ivar executor: manages parallel execution for the system
    :ivar graph: the internal graph data structure of the MD system

    :vartype exo_vars: list[BaseRV]
    :vartype coupling_vars: list[BaseRV]
    :vartype refine_level: int
    :vartype build_metrics: dict
    :vartype root_dir: str
    :vartype log_file: str
    :vartype executor: Executor
    :vartype graph: nx.DiGraph
    """

    def __init__(self, components: list[ComponentSpec] | ComponentSpec, exo_vars: list[BaseRV] | BaseRV,
                 coupling_vars: list[BaseRV] | BaseRV, est_bds: int = 0, save_dir: str | Path = None,
                 executor: Executor = None, stdout: bool = True, init_surr: bool = True):
        """Construct the MD system surrogate.

        !!! Warning
            Component models should always use coupling variables in the order they appear in the system-level
            `coupling_vars`.

        :param components: list of components in the MD system (using the ComponentSpec class)
        :param exo_vars: list of system-level exogenous/external inputs
        :param coupling_vars: list of all coupling variables (including all system-level outputs)
        :param est_bds: number of samples to estimate coupling variable bounds, do nothing if 0
        :param save_dir: root directory for all build products (.log, .pkl, .json, etc.), won't save if None
        :param executor: an instance of a `concurrent.futures.Executor`, use to iterate new candidates in parallel
        :param stdout: whether to log to console
        :param init_surr: whether to initialize the surrogate immediately when constructing
        """
        # Setup root save directory
        if save_dir is not None:
            timestamp = datetime.datetime.now(tz=timezone.utc).isoformat().split('.')[0].replace(':', '.')
            save_dir = Path(save_dir) / ('amisc_' + timestamp)
            os.mkdir(save_dir)
            self.root_dir = str(save_dir.resolve())
            os.mkdir(Path(self.root_dir) / 'sys')
            os.mkdir(Path(self.root_dir) / 'components')
            fname = timestamp + 'UTC_sys.log'
            self.log_file = str((Path(self.root_dir) / fname).resolve())
        else:
            self.root_dir = None
            self.log_file = None
        self.logger = get_logger(self.__class__.__name__, log_file=self.log_file, stdout=stdout)
        self.executor = executor

        # Store system info in a directed graph data structure
        self.graph = nx.DiGraph()
        self.exo_vars = copy.deepcopy(exo_vars) if isinstance(exo_vars, list) else [exo_vars]
        self.coupling_vars = copy.deepcopy(coupling_vars) if isinstance(coupling_vars, list) else [coupling_vars]
        self.refine_level = 0
        self.build_metrics = dict()     # Save refinement error metrics during training

        # Construct graph nodes
        components = [components] if not isinstance(components, list) else components
        for k, comp in enumerate(components):
            if comp['name'] == '':
                comp['name'] = f'Component {k}'
        Nk = len(components)
        nodes = {comp['name']: comp for comp in components}  # work-around since self.graph.nodes is not built yet
        for k in range(Nk):
            # Add the component as a str() node, with attributes specifying details of the surrogate
            comp_dict = components[k]
            indices, surr = self._build_component(comp_dict, nodes=nodes)
            self.graph.add_node(comp_dict['name'], surrogate=surr, is_computed=False, **indices)

        # Connect all neighbor nodes
        for node, node_obj in self.graph.nodes.items():
            for neighbor in node_obj['local_in']:
                self.graph.add_edge(neighbor, node)

        # Estimate coupling variable bounds
        if est_bds > 0:
            self._estimate_coupling_bds(est_bds)

        # Init system with most coarse fidelity indices in each component
        if init_surr:
            self.init_system()
        self.save_to_file('sys_init.pkl')

    def _build_component(self, component: ComponentSpec, nodes=None) -> tuple[dict, ComponentSurrogate]:
        """Build and return a `ComponentSurrogate` from a `dict` that describes the component model/connections.

        :param component: specifies details of a component (see `ComponentSpec`)
        :param nodes: `dict` of `{node: node_attributes}`, defaults to `self.graph.nodes`
        :returns: `connections, surr`: a `dict` of all connection indices and the `ComponentSurrogate` object
        """
        nodes = self.graph.nodes if nodes is None else nodes
        kwargs = component.get('model_kwargs', {})
        kwargs = {} if kwargs is None else kwargs

        # Set up defaults if this is a trivial one component system
        exo_in = component.get('exo_in', None)
        coupling_in = component.get('coupling_in', None)
        coupling_out = component.get('coupling_out', None)
        if len(nodes) == 1:
            exo_in = list(np.arange(0, len(self.exo_vars)))
            coupling_in = []
            coupling_out = list(np.arange(0, len(self.coupling_vars)))
        else:
            exo_in = [] if exo_in is None else exo_in
            coupling_in = [] if coupling_in is None else coupling_in
            coupling_out = [] if coupling_out is None else coupling_out
        exo_in = [exo_in] if not isinstance(exo_in, list) else exo_in
        coupling_in = [coupling_in] if not isinstance(coupling_in, list | dict) else coupling_in
        component['coupling_out'] = [coupling_out] if not isinstance(coupling_out, list) else coupling_out

        # Raise an error if all inputs or all outputs are empty
        if len(exo_in) + len(coupling_in) == 0:
            raise ValueError(f'Component {component["name"]} has no inputs! Please specify inputs in '
                             f'"exo_in" or "coupling_in".')
        if len(component['coupling_out']) == 0:
            raise ValueError(f'Component {component["name"]} has no outputs! Please specify outputs in '
                             f'"coupling_out".')

        # Get exogenous input indices (might already be a list of ints, otherwise convert list of vars to indices)
        if len(exo_in) > 0:
            if isinstance(exo_in[0], str | BaseRV):
                exo_in = [self.exo_vars.index(var) for var in exo_in]

        # Get global coupling output indices for all nodes (convert list of vars to list of indices if necessary)
        global_out = {}
        for node, node_obj in nodes.items():
            node_use = node_obj if node != component.get('name') else component
            coupling_out = node_use.get('coupling_out', None)
            coupling_out = [] if coupling_out is None else coupling_out
            coupling_out = [coupling_out] if not isinstance(coupling_out, list) else coupling_out
            global_out[node] = [self.coupling_vars.index(var) for var in coupling_out] if isinstance(
                coupling_out[0], str | BaseRV) else coupling_out

        # Refactor coupling inputs into both local and global index formats
        local_in = dict()  # e.g. {'Cathode': [0, 1, 2], 'Thruster': [0,], etc...}
        global_in = list()  # e.g. [0, 2, 4, 5, 6]
        if isinstance(coupling_in, dict):
            # If already a dict, get local connection indices from each neighbor
            for node, connections in coupling_in.items():
                conn_list = [connections] if not isinstance(connections, list) else connections
                if isinstance(conn_list[0], str | BaseRV):
                    global_ind = [self.coupling_vars.index(var) for var in conn_list]
                    local_in[node] = sorted([global_out[node].index(i) for i in global_ind])
                else:
                    local_in[node] = sorted(conn_list)

            # Convert to global coupling indices
            for node, local_idx in local_in.items():
                global_in.extend([global_out[node][i] for i in local_idx])
            global_in = sorted(global_in)
        else:
            # Otherwise, convert a list of global indices or vars into a dict of local indices
            if len(coupling_in) > 0:
                if isinstance(coupling_in[0], str | BaseRV):
                    coupling_in = [self.coupling_vars.index(var) for var in coupling_in]
            global_in = sorted(coupling_in)
            for node, node_obj in nodes.items():
                if node != component['name']:
                    l = list()
                    for i in global_in:
                        try:
                            l.append(global_out[node].index(i))
                        except ValueError:
                            pass
                    if l:
                        local_in[node] = sorted(l)

        # Store all connection indices for this component
        connections = dict(exo_in=exo_in, local_in=local_in, global_in=global_in,
                           global_out=global_out.get(component.get('name')))

        # Set up a component output save directory
        if component.get('save_output', False) and self.root_dir is not None:
            output_dir = str((Path(self.root_dir) / 'components' / component['name']).resolve())
            if not Path(output_dir).is_dir():
                os.mkdir(output_dir)
            kwargs['output_dir'] = output_dir
        else:
            if kwargs.get('output_dir', None) is not None:
                kwargs['output_dir'] = None

        # Initialize a new component surrogate
        surr_class = component.get('surrogate', 'lagrange')
        if isinstance(surr_class, str):
            match surr_class:
                case 'lagrange':
                    surr_class = SparseGridSurrogate
                case 'analytical':
                    surr_class = AnalyticalSurrogate
                case other:
                    raise NotImplementedError(f"Surrogate type '{surr_class}' is not known at this time.")

        # Check for an override of model fidelity indices (to enable just single-fidelity evaluation)
        if kwargs.get('hf_override', False):
            truth_alpha, max_alpha = (), ()
            kwargs['hf_override'] = component['truth_alpha']    # Pass in the truth alpha indices as a kwarg to model
        else:
            truth_alpha, max_alpha = component['truth_alpha'], component['max_alpha']
        max_beta = component.get('max_beta')
        truth_alpha = (truth_alpha,) if isinstance(truth_alpha, int) else truth_alpha
        max_alpha = (max_alpha,) if isinstance(max_alpha, int) else max_alpha
        max_beta = (max_beta,) if isinstance(max_beta, int) else max_beta

        # Assumes input ordering is exogenous vars + sorted coupling vars
        x_vars = [self.exo_vars[i] for i in exo_in] + [self.coupling_vars[i] for i in global_in]
        surr = surr_class(x_vars, component['model'], truth_alpha=truth_alpha, max_alpha=max_alpha,
                          max_beta=max_beta, executor=self.executor, log_file=self.log_file,
                          model_args=component.get('model_args'), model_kwargs=kwargs)
        return connections, surr

    def swap_component(self, component: ComponentSpec, exo_add: BaseRV | list[BaseRV] = None,
                       exo_remove: IndicesRV = None, qoi_add: BaseRV | list[BaseRV] = None,
                       qoi_remove: IndicesRV = None):
        """Swap a new component into the system, updating all connections/inputs.

        !!! Warning "Beta feature, proceed with caution"
            If you are swapping a new component in, you cannot remove any inputs that are expected by other components,
            including the coupling variables output by the current model.

        :param component: specs of new component model (must replace an existing component with matching `name`)
        :param exo_add: variables to add to system exogenous inputs (will be appended to end)
        :param exo_remove: indices of system exogenous inputs to delete (can't be shared by other components)
        :param qoi_add: system output QoIs to add (will be appended to end of `coupling_vars`)
        :param qoi_remove: indices of system `coupling_vars` to delete (can't be shared by other components)
        """
        # Delete system exogenous inputs
        if exo_remove is None:
            exo_remove = []
        exo_remove = [exo_remove] if not isinstance(exo_remove, list) else exo_remove
        exo_remove = [self.exo_vars.index(var) for var in exo_remove] if exo_remove and isinstance(
            exo_remove[0], str | BaseRV) else exo_remove

        exo_remove = sorted(exo_remove)
        for j, exo_var_idx in enumerate(exo_remove):
            # Adjust exogenous indices for all components to account for deleted system inputs
            for node, node_obj in self.graph.nodes.items():
                if node != component['name']:
                    for i, idx in enumerate(node_obj['exo_in']):
                        if idx == exo_var_idx:
                            error_msg = f"Can't delete system exogenous input at idx {exo_var_idx}, since it is " \
                                        f"shared by component '{node}'."
                            self.logger.error(error_msg)
                            raise ValueError(error_msg)
                        if idx > exo_var_idx:
                            node_obj['exo_in'][i] -= 1

            # Need to update the remaining delete indices by -1 to account for each sequential deletion
            del self.exo_vars[exo_var_idx]
            for i in range(j+1, len(exo_remove)):
                exo_remove[i] -= 1

        # Append any new exogenous inputs to the end
        if exo_add is not None:
            exo_add = [exo_add] if not isinstance(exo_add, list) else exo_add
            self.exo_vars.extend(exo_add)

        # Delete system qoi outputs (if not shared by other components)
        qoi_remove = sorted(self._get_qoi_ind(qoi_remove))
        for j, qoi_idx in enumerate(qoi_remove):
            # Adjust coupling indices for all components to account for deleted system outputs
            for node, node_obj in self.graph.nodes.items():
                if node != component['name']:
                    for i, idx in enumerate(node_obj['global_in']):
                        if idx == qoi_idx:
                            error_msg = f"Can't delete system QoI at idx {qoi_idx}, since it is an input to " \
                                        f"component '{node}'."
                            self.logger.error(error_msg)
                            raise ValueError(error_msg)
                        if idx > qoi_idx:
                            node_obj['global_in'][i] -= 1

                    for i, idx in enumerate(node_obj['global_out']):
                        if idx > qoi_idx:
                            node_obj['global_out'][i] -= 1

            # Need to update the remaining delete indices by -1 to account for each sequential deletion
            del self.coupling_vars[qoi_idx]
            for i in range(j+1, len(qoi_remove)):
                qoi_remove[i] -= 1

        # Append any new system QoI outputs to the end
        if qoi_add is not None:
            qoi_add = [qoi_add] if not isinstance(qoi_add, list) else qoi_add
            self.coupling_vars.extend(qoi_add)

        # Build and initialize the new component surrogate
        indices, surr = self._build_component(component)
        surr.init_coarse()

        # Make changes to adj matrix if coupling inputs changed
        prev_neighbors = list(self.graph.nodes[component['name']]['local_in'].keys())
        new_neighbors = list(indices['local_in'].keys())
        for neighbor in new_neighbors:
            if neighbor not in prev_neighbors:
                self.graph.add_edge(neighbor, component['name'])
            else:
                prev_neighbors.remove(neighbor)
        for neighbor in prev_neighbors:
            self.graph.remove_edge(neighbor, component['name'])

        self.logger.info(f"Swapped component '{component['name']}'.")
        nx.set_node_attributes(self.graph, {component['name']: {'exo_in': indices['exo_in'], 'local_in':
                                                                indices['local_in'], 'global_in': indices['global_in'],
                                                                'global_out': indices['global_out'],
                                                                'surrogate': surr, 'is_computed': False}})

    def insert_component(self, component: ComponentSpec, exo_add: BaseRV | list[BaseRV] = None,
                         qoi_add: BaseRV | list[BaseRV] = None):
        """Insert a new component into the system.

        :param component: specs of new component model
        :param exo_add: variables to add to system exogenous inputs (will be appended to end of `exo_vars`)
        :param qoi_add: system output QoIs to add (will be appended to end of `coupling_vars`)
        """
        if exo_add is not None:
            exo_add = [exo_add] if not isinstance(exo_add, list) else exo_add
            self.exo_vars.extend(exo_add)
        if qoi_add is not None:
            qoi_add = [qoi_add] if not isinstance(qoi_add, list) else qoi_add
            self.coupling_vars.extend(qoi_add)

        indices, surr = self._build_component(component)
        surr.init_coarse()
        self.graph.add_node(component['name'], surrogate=surr, is_computed=False, **indices)

        # Add graph edges
        neighbors = list(indices['local_in'].keys())
        for neighbor in neighbors:
            self.graph.add_edge(neighbor, component['name'])
        self.logger.info(f"Inserted component '{component['name']}'.")

    def _save_on_error(func):
        """Gracefully exit and save `SystemSurrogate` on any errors."""
        @functools.wraps(func)
        def wrap(self, *args, **kwargs):
            try:
                return func(self, *args, **kwargs)
            except:
                self.save_to_file('sys_error.pkl')
                self.logger.critical(f'An error occurred during execution of {func.__name__}. Saving '
                                     f'SystemSurrogate object to sys_error.pkl', exc_info=True)
                self.logger.info(f'Final system surrogate on exit: \n {self}')
                raise
        return wrap
    _save_on_error = staticmethod(_save_on_error)

    @_save_on_error
    def init_system(self):
        """Add the coarsest multi-index to each component surrogate."""
        self._print_title_str('Initializing all component surrogates')
        for node, node_obj in self.graph.nodes.items():
            node_obj['surrogate'].init_coarse()
            # for alpha, beta in list(node_obj['surrogate'].candidate_set):
            #     # Add one refinement in each input dimension to initialize
            #     node_obj['surrogate'].activate_index(alpha, beta)
            self.logger.info(f"Initialized component '{node}'.")

    @_save_on_error
    def fit(self, qoi_ind: IndicesRV = None, num_refine: int = 100, max_iter: int = 20, max_tol: float = 1e-3,
            max_runtime: float = 1, save_interval: int = 0, update_bounds: bool = True, test_set: dict = None,
            n_jobs: int = 1):
        """Train the system surrogate adaptively by iterative refinement until an end condition is met.

        :param qoi_ind: list of system QoI variables to focus refinement on, use all QoI if not specified
        :param num_refine: number of samples of exogenous inputs to compute error indicators on
        :param max_iter: the maximum number of refinement steps to take
        :param max_tol: the max allowable value in relative L2 error to achieve
        :param max_runtime: the maximum wall clock time (hr) to run refinement for (will go until all models finish)
        :param save_interval: number of refinement steps between each progress save, none if 0
        :param update_bounds: whether to continuously update coupling variable bounds during refinement
        :param test_set: `dict(xt=(Nt, x_dim), yt=(Nt, y_dim)` to show convergence of surrogate to the truth model
        :param n_jobs: number of cpu workers for computing error indicators (on master MPI task), 1=sequential
        """
        qoi_ind = self._get_qoi_ind(qoi_ind)
        Nqoi = len(qoi_ind)
        max_iter = self.refine_level + max_iter
        curr_error = np.inf
        t_start = time.time()
        test_stats, xt, yt, t_fig, t_ax = None, None, None, None, None

        # Record of (error indicator, component, alpha, beta, num_evals, total added cost (s)) for each iteration
        train_record = self.build_metrics.get('train_record', [])
        if test_set is not None:
            xt, yt = test_set['xt'], test_set['yt']
        xt, yt = self.build_metrics.get('xt', xt), self.build_metrics.get('yt', yt)  # Overrides test set param

        # Track convergence progress on a test set and on the max error indicator
        err_fig, err_ax = plt.subplots()
        if xt is not None and yt is not None:
            self.build_metrics['xt'] = xt
            self.build_metrics['yt'] = yt
            if self.build_metrics.get('test_stats') is not None:
                test_stats = self.build_metrics.get('test_stats')
            else:
                # Get initial perf metrics, (2, Nqoi)
                test_stats = np.expand_dims(self.get_test_metrics(xt, yt, qoi_ind=qoi_ind), axis=0)
            t_fig, t_ax = plt.subplots(1, Nqoi) if Nqoi > 1 else plt.subplots()

        # Set up a parallel pool of workers, sequential if n_jobs=1
        with Parallel(n_jobs=n_jobs, verbose=0) as ppool:
            while True:
                # Check all end conditions
                if self.refine_level >= max_iter:
                    self._print_title_str(f'Termination criteria reached: Max iteration {self.refine_level}/{max_iter}')
                    break
                if curr_error == -np.inf:
                    self._print_title_str(f'Termination criteria reached: No candidates left to refine')
                    break
                if curr_error < max_tol:
                    self._print_title_str(f'Termination criteria reached: L2 error {curr_error} < tol {max_tol}')
                    break
                if ((time.time() - t_start)/3600.0) >= max_runtime:
                    actual = datetime.timedelta(seconds=time.time()-t_start)
                    target = datetime.timedelta(seconds=max_runtime*3600)
                    self._print_title_str(f'Termination criteria reached: runtime {str(actual)} > {str(target)}')
                    break

                # Refine surrogate and save progress
                refine_res = self.refine(qoi_ind=qoi_ind, num_refine=num_refine, update_bounds=update_bounds,
                                         ppool=ppool)
                curr_error = refine_res[0]
                if save_interval > 0 and self.refine_level % save_interval == 0:
                    self.save_to_file(f'sys_iter_{self.refine_level}.pkl')

                # Plot progress of error indicator
                train_record.append(refine_res)
                error_record = [res[0] for res in train_record]
                self.build_metrics['train_record'] = train_record
                err_ax.clear(); err_ax.grid(); err_ax.plot(error_record, '-k')
                ax_default(err_ax, 'Iteration', r'Relative $L_2$ error indicator', legend=False)
                err_ax.set_yscale('log')
                if self.root_dir is not None:
                    err_fig.savefig(str(Path(self.root_dir) / 'error_indicator.png'), dpi=300, format='png')

                # Plot progress on test set
                if xt is not None and yt is not None:
                    stats = self.get_test_metrics(xt, yt, qoi_ind=qoi_ind)
                    test_stats = np.concatenate((test_stats, stats[np.newaxis, ...]), axis=0)
                    for i in range(Nqoi):
                        ax = t_ax if Nqoi == 1 else t_ax[i]
                        ax.clear(); ax.grid(); ax.set_yscale('log')
                        ax.plot(test_stats[:, 1, i], '-k')
                        ax.set_title(self.coupling_vars[qoi_ind[i]].to_tex(units=True))
                        ax_default(ax, 'Iteration', r'Relative $L_2$ error', legend=False)
                    t_fig.set_size_inches(3.5*Nqoi, 3.5)
                    t_fig.tight_layout()
                    if self.root_dir is not None:
                        t_fig.savefig(str(Path(self.root_dir) / 'test_set.png'), dpi=300, format='png')
                    self.build_metrics['test_stats'] = test_stats

        self.save_to_file(f'sys_final.pkl')
        self.logger.info(f'Final system surrogate: \n {self}')

    def get_allocation(self, idx: int = None):
        """Get a breakdown of cost allocation up to a certain iteration number during training (starting at 1).

        :param idx: the iteration number to get allocation results for (defaults to last refinement step)
        :returns: `cost_alloc, offline_alloc, cost_cum` - the cost alloc per node/fidelity and cumulative training cost
        """
        if idx is None:
            idx = self.refine_level
        if idx > self.refine_level:
            raise ValueError(f'Specified index: {idx} is greater than the max training level of {self.refine_level}')

        cost_alloc = dict()     # Cost allocation per node and model fidelity
        cost_cum = [0.0]        # Cumulative cost allocation during training

        # Add initialization costs for each node
        for node, node_obj in self.graph.nodes.items():
            surr = node_obj['surrogate']
            base_alpha = (0,) * len(surr.truth_alpha)
            base_beta = (0,) * (len(surr.max_refine) - len(surr.truth_alpha))
            base_cost = surr.get_cost(base_alpha, base_beta)
            cost_alloc[node] = dict()
            if base_cost > 0:
                cost_alloc[node][str(base_alpha)] = np.array([1, float(base_cost)])
                cost_cum[0] += float(base_cost)

        # Add cumulative training costs
        for i in range(idx):
            err_indicator, node, alpha, beta, num_evals, cost = self.build_metrics['train_record'][i]
            if cost_alloc[node].get(str(alpha), None) is None:
                cost_alloc[node][str(alpha)] = np.zeros(2)  # (num model evals, total cpu_time cost)
            cost_alloc[node][str(alpha)] += [round(num_evals), float(cost)]
            cost_cum.append(float(cost))

        # Get summary of total offline costs spent building search candidates (i.e. training overhead)
        offline_alloc = dict()
        for node, node_obj in self.graph.nodes.items():
            surr = node_obj['surrogate']
            offline_alloc[node] = dict()
            for alpha, beta in surr.candidate_set:
                if offline_alloc[node].get(str(alpha), None) is None:
                    offline_alloc[node][str(alpha)] = np.zeros(2)   # (num model evals, total cpu_time cost)
                added_cost = surr.get_cost(alpha, beta)
                base_cost = surr.get_sub_surrogate(alpha, beta).model_cost
                offline_alloc[node][str(alpha)] += [round(added_cost/base_cost), float(added_cost)]

        return cost_alloc, offline_alloc, np.cumsum(cost_cum)

    def get_test_metrics(self, xt: np.ndarray, yt: np.ndarray, qoi_ind: IndicesRV = None,
                         training: bool = True) -> np.ndarray:
        """Get relative L2 error metric over a test set.

        :param xt: `(Nt, x_dim)` random test set of inputs
        :param yt: `(Nt, y_dim)` random test set outputs
        :param qoi_ind: list of indices of QoIs to get metrics for
        :param training: whether to evaluate the surrogate in training or evaluation mode
        :returns: `stats` - `(2, Nqoi)` array &rarr; `[num_candidates, rel_L2_error]` for each QoI
        """
        qoi_ind = self._get_qoi_ind(qoi_ind)
        ysurr = self(xt, training=training)
        ysurr = ysurr[:, qoi_ind]
        yt = yt[:, qoi_ind]
        with np.errstate(divide='ignore', invalid='ignore'):
            rel_l2_err = np.sqrt(np.mean((yt - ysurr) ** 2, axis=0)) / np.sqrt(np.mean(yt ** 2, axis=0))
            rel_l2_err = np.nan_to_num(rel_l2_err, posinf=np.nan, neginf=np.nan, nan=np.nan)
        num_cands = 0
        for node, node_obj in self.graph.nodes.items():
            num_cands += len(node_obj['surrogate'].index_set) + len(node_obj['surrogate'].candidate_set)

        # Get test stats for each QoI
        stats = np.zeros((2, yt.shape[-1]))
        self.logger.debug(f'{"QoI idx": >10} {"Iteration": >10} {"len(I_k)": >10} {"Relative L2": >15}')
        for i in range(yt.shape[-1]):
            stats[:, i] = np.array([num_cands, rel_l2_err[i]])
            self.logger.debug(f'{i: 10d} {self.refine_level: 10d} {num_cands: 10d} {rel_l2_err[i]: 15.5f}')

        return stats

    def _get_qoi_ind(self, qoi_ind: IndicesRV) -> list[int]:
        """Small helper to make sure QoI indices are a list of integers."""
        if qoi_ind is None:
            qoi_ind = list(np.arange(0, len(self.coupling_vars)))
        qoi_ind = [qoi_ind] if not isinstance(qoi_ind, list) else qoi_ind
        qoi_ind = [self.coupling_vars.index(var) for var in qoi_ind] if qoi_ind and isinstance(
            qoi_ind[0], str | BaseRV) else qoi_ind

        return qoi_ind

    def refine(self, qoi_ind: IndicesRV = None, num_refine: int = 100, update_bounds: bool = True,
               ppool: Parallel = None) -> tuple:
        """Find and refine the component surrogate with the largest error on system-level QoI.

        :param qoi_ind: indices of system QoI to focus surrogate refinement on, use all QoI if not specified
        :param num_refine: number of samples of exogenous inputs to compute error indicators on
        :param update_bounds: whether to continuously update coupling variable bounds
        :param ppool: a `Parallel` instance from `joblib` to compute error indicators in parallel, None=sequential
        :returns refine_res: a tuple of `(error_indicator, component, node_star, alpha_star, beta_star, N, cost)`
                             indicating the chosen candidate index and incurred cost
        """
        self._print_title_str(f'Refining system surrogate: iteration {self.refine_level + 1}')
        set_loky_pickler('dill')    # Dill can serialize 'self' for parallel workers
        temp_exc = self.executor    # It can't serialize an executor though, so must save this temporarily
        self.set_executor(None)
        qoi_ind = self._get_qoi_ind(qoi_ind)

        # Compute entire integrated-surrogate on a random test set for global system QoI error estimation
        x_exo = self.sample_inputs((num_refine,))
        y_curr = self(x_exo, training=True)
        y_min, y_max = None, None
        if update_bounds:
            y_min = np.min(y_curr, axis=0, keepdims=True)  # (1, ydim)
            y_max = np.max(y_curr, axis=0, keepdims=True)  # (1, ydim)

        # Find the candidate surrogate with the largest error indicator
        error_max, error_indicator = -np.inf, -np.inf
        node_star, alpha_star, beta_star, l2_star, cost_star = None, None, None, -np.inf, 0
        for node, node_obj in self.graph.nodes.items():
            self.logger.info(f"Estimating error for component '{node}'...")
            candidates = node_obj['surrogate'].candidate_set.copy()

            def compute_error(alpha, beta):
                # Helper function for computing error indicators for a given candidate (alpha, beta)
                index_set = node_obj['surrogate'].index_set.copy()
                index_set.append((alpha, beta))
                y_cand = self(x_exo, training=True, index_set={node: index_set})
                ymin = np.min(y_cand, axis=0, keepdims=True)
                ymax = np.max(y_cand, axis=0, keepdims=True)
                error = y_cand[:, qoi_ind] - y_curr[:, qoi_ind]
                rel_l2 = np.sqrt(np.nanmean(error ** 2, axis=0)) / np.sqrt(np.nanmean(y_cand[:, qoi_ind] ** 2, axis=0))
                rel_l2 = np.nan_to_num(rel_l2, nan=np.nan, posinf=np.nan, neginf=np.nan)
                delta_error = np.nanmax(rel_l2)  # Max relative L2 error over all system QoIs
                delta_work = max(1, node_obj['surrogate'].get_cost(alpha, beta))  # Cpu time (s)

                return ymin, ymax, delta_error, delta_work

            if len(candidates) > 0:
                ret = ppool(delayed(compute_error)(alpha, beta) for alpha, beta in candidates) if ppool is not None \
                    else [compute_error(alpha, beta) for alpha, beta in candidates]

                for i, (ymin, ymax, d_error, d_work) in enumerate(ret):
                    if update_bounds:
                        y_min = np.min(np.concatenate((y_min, ymin), axis=0), axis=0, keepdims=True)
                        y_max = np.max(np.concatenate((y_max, ymax), axis=0), axis=0, keepdims=True)
                    alpha, beta = candidates[i]
                    error_indicator = d_error / d_work
                    self.logger.info(f"Candidate multi-index: {(alpha, beta)}. Error indicator: "
                                     f"{error_indicator}. L2 error: {d_error}")

                    if error_indicator > error_max:
                        error_max = error_indicator
                        node_star, alpha_star, beta_star, l2_star, cost_star = node, alpha, beta, d_error, d_work
            else:
                self.logger.info(f"Component '{node}' has no available candidates left!")

        # Update all coupling variable ranges
        if update_bounds:
            for i in range(y_curr.shape[-1]):
                self._update_coupling_bds(i, (y_min[0, i], y_max[0, i]))

        # Add the chosen multi-index to the chosen component
        self.set_executor(temp_exc)
        if node_star is not None:
            self.logger.info(f"Candidate multi-index {(alpha_star, beta_star)} chosen for component '{node_star}'")
            self.graph.nodes[node_star]['surrogate'].activate_index(alpha_star, beta_star)
            self.refine_level += 1
            num_evals = round(cost_star / self[node_star].get_sub_surrogate(alpha_star, beta_star).model_cost)
        else:
            self.logger.info(f"No candidates left for refinement, iteration: {self.refine_level}")
            num_evals = 0

        return l2_star, node_star, alpha_star, beta_star, num_evals, cost_star

    def predict(self, x: np.ndarray | float, max_fpi_iter: int = 100, anderson_mem: int = 10, fpi_tol: float = 1e-10,
                use_model: str | tuple | dict = None, model_dir: str | Path = None, verbose: bool = False,
                training: bool = False, index_set: dict[str: IndexSet] = None, qoi_ind: IndicesRV = None) -> np.ndarray:
        """Evaluate the system surrogate at exogenous inputs `x`.

        !!! Warning
            You can use this function to predict outputs for your MD system using the full-order models rather than the
            surrogate, by specifying `use_model`. This is convenient because `SystemSurrogate` manages all the
            coupled information flow between models automatically. However, it is *highly* recommended to not use
            the full model if your system contains feedback loops. The FPI nonlinear solver would be infeasible using
            anything more computationally demanding than the surrogate.

        :param x: `(..., x_dim)` the points to get surrogate predictions for
        :param max_fpi_iter: the limit on convergence for the fixed-point iteration routine
        :param anderson_mem: hyperparameter for tuning the convergence of FPI with anderson acceleration
        :param fpi_tol: tolerance limit for convergence of fixed-point iteration
        :param use_model: 'best'=highest-fidelity, 'worst'=lowest-fidelity, tuple=specific fidelity, None=surrogate,
                           specify a `dict` of the above to assign different model fidelities for diff components
        :param model_dir: directory to save model outputs if `use_model` is specified
        :param verbose: whether to print out iteration progress during execution
        :param training: whether to call the system surrogate in training or evaluation mode, ignored if `use_model`
        :param index_set: `dict(node=[indices])` to override default index set for a node (only useful for parallel)
        :param qoi_ind: list of qoi indices to return, defaults to returning all system `coupling_vars`
        :returns y: `(..., y_dim)` the surrogate approximation of the system QoIs
        """
        # Allocate space for all system outputs (just save all coupling vars)
        x = np.atleast_1d(x)
        ydim = len(self.coupling_vars)
        y = np.zeros(x.shape[:-1] + (ydim,))
        valid_idx = ~np.isnan(x[..., 0])  # Keep track of valid samples (set to False if FPI fails)
        t1 = 0
        output_dir = None

        # Interpret which model fidelities to use for each component (if specified)
        if use_model is not None:
            if not isinstance(use_model, dict):
                use_model = {node: use_model for node in self.graph.nodes}  # use same for each component
        else:
            use_model = {node: None for node in self.graph.nodes}
        use_model = {node: use_model.get(node, None) for node in self.graph.nodes}

        # Initialize all components
        for node, node_obj in self.graph.nodes.items():
            node_obj['is_computed'] = False

        # Convert system into DAG by grouping strongly-connected-components
        dag = nx.condensation(self.graph)

        # Compute component models in topological order
        for supernode in nx.topological_sort(dag):
            scc = [n for n in dag.nodes[supernode]['members']]

            # Compute single component feedforward output (no FPI needed)
            if len(scc) == 1:
                if verbose:
                    self.logger.info(f"Running component '{scc[0]}'...")
                    t1 = time.time()

                # Gather inputs
                node_obj = self.graph.nodes[scc[0]]
                exo_inputs = x[..., node_obj['exo_in']]
                # for comp_name in node_obj['local_in']:
                #     assert self.graph.nodes[comp_name]['is_computed']
                coupling_inputs = y[..., node_obj['global_in']]
                comp_input = np.concatenate((exo_inputs, coupling_inputs), axis=-1)  # (..., xdim)

                # Compute outputs
                indices = index_set.get(scc[0], None) if index_set is not None else None
                if model_dir is not None:
                    output_dir = Path(model_dir) / scc[0]
                    os.mkdir(output_dir)
                comp_output = node_obj['surrogate'](comp_input[valid_idx, :], use_model=use_model.get(scc[0]),
                                                    model_dir=output_dir, training=training, index_set=indices)
                for local_i, global_i in enumerate(node_obj['global_out']):
                    y[valid_idx, global_i] = comp_output[..., local_i]
                node_obj['is_computed'] = True

                if verbose:
                    self.logger.info(f"Component '{scc[0]}' completed. Runtime: {time.time() - t1} s")

            # Handle FPI for SCCs with more than one component
            else:
                # Set the initial guess for all coupling vars (middle of domain)
                coupling_bds = [rv.bounds() for rv in self.coupling_vars]
                x_couple = np.array([(bds[0] + bds[1]) / 2 for bds in coupling_bds])
                x_couple = np.broadcast_to(x_couple, x.shape[:-1] + x_couple.shape).copy()

                adj_nodes = []
                fpi_idx = set()
                for node in scc:
                    for comp_name, local_idx in self.graph.nodes[node]['local_in'].items():
                        # Track the global idx of all coupling vars that need FPI
                        if comp_name in scc:
                            fpi_idx.update([self.graph.nodes[comp_name]['global_out'][idx] for idx in local_idx])

                        # Override coupling vars from components outside the scc (should already be computed)
                        if comp_name not in scc and comp_name not in adj_nodes:
                            # assert self.graph.nodes[comp_name]['is_computed']
                            global_idx = self.graph.nodes[comp_name]['global_out']
                            x_couple[..., global_idx] = y[..., global_idx]
                            adj_nodes.append(comp_name)  # Only need to do this once for each adj component
                x_couple_next = x_couple.copy()
                fpi_idx = sorted(fpi_idx)

                # Main FPI loop
                if verbose:
                    self.logger.info(f"Initializing FPI for SCC {scc} ...")
                    t1 = time.time()
                k = 0
                residual_hist = None
                x_hist = None
                while True:
                    for node in scc:
                        # Gather inputs from exogenous and coupling sources
                        node_obj = self.graph.nodes[node]
                        exo_inputs = x[..., node_obj['exo_in']]
                        coupling_inputs = x_couple[..., node_obj['global_in']]
                        comp_input = np.concatenate((exo_inputs, coupling_inputs), axis=-1)     # (..., xdim)

                        # Compute component outputs (just don't do this FPI with the real models, please..)
                        indices = index_set.get(node, None) if index_set is not None else None
                        comp_output = node_obj['surrogate'](comp_input[valid_idx, :], use_model=use_model.get(node),
                                                            model_dir=None, training=training, index_set=indices)
                        global_idx = node_obj['global_out']
                        for local_i, global_i in enumerate(global_idx):
                            x_couple_next[valid_idx, global_i] = comp_output[..., local_i]
                            # Can't splice valid_idx with global_idx for some reason, have to loop over global_idx here

                    # Compute residual and check end conditions
                    residual = np.expand_dims(x_couple_next[..., fpi_idx] - x_couple[..., fpi_idx], axis=-1)
                    max_error = np.max(np.abs(residual[valid_idx, :, :]))
                    if verbose:
                        self.logger.info(f'FPI iter: {k}. Max residual: {max_error}. Time: {time.time() - t1} s')
                    if max_error <= fpi_tol:
                        if verbose:
                            self.logger.info(f'FPI converged for SCC {scc} in {k} iterations with {max_error} < tol '
                                             f'{fpi_tol}. Final time: {time.time() - t1} s')
                        break
                    if k >= max_fpi_iter:
                        self.logger.warning(f'FPI did not converge in {max_fpi_iter} iterations for SCC {scc}: '
                                            f'{max_error} > tol {fpi_tol}. Some samples will be returned as NaN.')
                        converged_idx = np.max(np.abs(residual), axis=(-1, -2)) <= fpi_tol
                        for idx in fpi_idx:
                            y[~converged_idx, idx] = np.nan
                        valid_idx = np.logical_and(valid_idx, converged_idx)
                        break

                    # Keep track of residual and x_couple histories
                    if k == 0:
                        residual_hist = residual.copy()                                 # (..., xdim, 1)
                        x_hist = np.expand_dims(x_couple_next[..., fpi_idx], axis=-1)   # (..., xdim, 1)
                        x_couple[:] = x_couple_next[:]
                        k += 1
                        continue  # skip anderson accel on first iteration

                    # Iterate with anderson acceleration (only iterate on samples that are not yet converged)
                    converged_idx = np.max(np.abs(residual), axis=(-1, -2)) <= fpi_tol
                    curr_idx = np.logical_and(valid_idx, ~converged_idx)
                    residual_hist = np.concatenate((residual_hist, residual), axis=-1)
                    x_hist = np.concatenate((x_hist, np.expand_dims(x_couple_next[..., fpi_idx], axis=-1)), axis=-1)
                    mk = min(anderson_mem, k)
                    Fk = residual_hist[curr_idx, :, k-mk:]                               # (..., xdim, mk+1)
                    C = np.ones(Fk.shape[:-2] + (1, mk + 1))
                    b = np.zeros(Fk.shape[:-2] + (len(fpi_idx), 1))
                    d = np.ones(Fk.shape[:-2] + (1, 1))
                    alpha = np.expand_dims(self._constrained_lls(Fk, b, C, d), axis=-3)   # (..., 1, mk+1, 1)
                    x_new = np.squeeze(x_hist[curr_idx, :, np.newaxis, -(mk+1):] @ alpha, axis=(-1, -2))
                    for local_i, global_i in enumerate(fpi_idx):
                        x_couple[curr_idx, global_i] = x_new[..., local_i]
                    k += 1

                # Save outputs of each component in SCC after convergence of FPI
                for node in scc:
                    global_idx = self.graph.nodes[node]['global_out']
                    for global_i in global_idx:
                        y[valid_idx, global_i] = x_couple_next[valid_idx, global_i]
                    self.graph.nodes[node]['is_computed'] = True

        # Return all component outputs (..., Nqoi), samples that didn't converge during FPI are left as np.nan
        qoi_ind = self._get_qoi_ind(qoi_ind)
        return y[..., qoi_ind]

    def __call__(self, *args, **kwargs):
        """Convenience wrapper to allow calling as `ret = SystemSurrogate(x)`."""
        return self.predict(*args, **kwargs)

    def _estimate_coupling_bds(self, num_est: int, anderson_mem: int = 10, fpi_tol: float = 1e-10,
                               max_fpi_iter: int = 100):
        """Estimate and set the coupling variable bounds.

        :param num_est: the number of samples of exogenous inputs to use
        :param anderson_mem: FPI hyperparameter (default is usually good)
        :param fpi_tol: floating point tolerance for FPI convergence
        :param max_fpi_iter: maximum number of FPI iterations
        """
        self._print_title_str('Estimating coupling variable bounds')
        x = self.sample_inputs((num_est,))
        y = self(x, use_model='best', verbose=True, anderson_mem=anderson_mem, fpi_tol=fpi_tol,
                 max_fpi_iter=max_fpi_iter)
        for i in range(len(self.coupling_vars)):
            lb = np.nanmin(y[:, i])
            ub = np.nanmax(y[:, i])
            self._update_coupling_bds(i, (lb, ub), init=True)

    def _update_coupling_bds(self, global_idx: int, bds: tuple, init: bool = False, buffer: float = 0.05):
        """Update coupling variable bounds.

        :param global_idx: global index of coupling variable to update
        :param bds: new bounds to update the current bounds with
        :param init: whether to set new bounds or update existing (default)
        :param buffer: fraction of domain length to buffer upper/lower bounds
        """
        offset = buffer * (bds[1] - bds[0])
        offset_bds = (bds[0] - offset, bds[1] + offset)
        coupling_bds = [rv.bounds() for rv in self.coupling_vars]
        new_bds = offset_bds if init else (min(coupling_bds[global_idx][0], offset_bds[0]),
                                           max(coupling_bds[global_idx][1], offset_bds[1]))
        self.coupling_vars[global_idx].update_bounds(*new_bds)

        # Iterate over all components and update internal coupling variable bounds
        for node_name, node_obj in self.graph.nodes.items():
            if global_idx in node_obj['global_in']:
                # Get the local index for this coupling variable within each component's inputs
                local_idx = len(node_obj['exo_in']) + node_obj['global_in'].index(global_idx)
                node_obj['surrogate'].update_input_bds(local_idx, new_bds)

    def sample_inputs(self, size: tuple | int, comp: str = 'System', use_pdf: bool = False,
                      nominal: dict[str: float] = None, constants: set[str] = None) -> np.ndarray:
        """Return samples of the inputs according to provided options.

        :param size: tuple or integer specifying shape or number of samples to obtain
        :param comp: which component to sample inputs for (defaults to full system exogenous inputs)
        :param use_pdf: whether to sample from each variable's pdf, defaults to random samples over input domain instead
        :param nominal: `dict(var_id=value)` of nominal values for params with relative uncertainty, also can use
                        to specify constant values for a variable listed in `constants`
        :param constants: set of param types to hold constant while sampling (i.e. calibration, design, etc.),
                          can also put a `var_id` string in here to specify a single variable to hold constant
        :returns x: `(*size, x_dim)` samples of the inputs for the given component/system
        """
        size = (size, ) if isinstance(size, int) else size
        if nominal is None:
            nominal = dict()
        if constants is None:
            constants = set()
        x_vars = self.exo_vars if comp == 'System' else self[comp].x_vars
        x = np.empty((*size, len(x_vars)))
        for i, var in enumerate(x_vars):
            # Set a constant value for this variable
            if var.param_type in constants or var in constants:
                x[..., i] = nominal.get(var, var.nominal)  # Defaults to variable's nominal value if not specified

            # Sample from this variable's pdf or randomly within its domain bounds (reject if outside bounds)
            else:
                lb, ub = var.bounds()
                x_sample = var.sample(size, nominal=nominal.get(var, None)) if use_pdf \
                    else var.sample_domain(size)
                good_idx = (x_sample < ub) & (x_sample > lb)
                num_reject = np.sum(~good_idx)

                while num_reject > 0:
                    new_sample = var.sample((num_reject,), nominal=nominal.get(var, None)) if use_pdf \
                        else var.sample_domain((num_reject,))
                    x_sample[~good_idx] = new_sample
                    good_idx = (x_sample < ub) & (x_sample > lb)
                    num_reject = np.sum(~good_idx)

                x[..., i] = x_sample

        return x

    def plot_slice(self, slice_idx: IndicesRV = None, qoi_idx: IndicesRV = None, show_surr: bool = True,
                   show_model: list = None, model_dir: str | Path = None, N: int = 50, nominal: dict[str: float] = None,
                   random_walk: bool = False, from_file: str | Path = None):
        """Helper function to plot 1d slices of the surrogate and/or model(s) over the inputs.

        :param slice_idx: list of exogenous input variables or indices to take 1d slices of
        :param qoi_idx: list of model output variables or indices to plot 1d slices of
        :param show_surr: whether to show the surrogate prediction
        :param show_model: also plot model predictions, list() of ['best', 'worst', tuple(alpha), etc.]
        :param model_dir: base directory to save model outputs (if specified)
        :param N: the number of points to take in the 1d slice
        :param nominal: `dict` of `str(var)->nominal` to use as constant value for all non-sliced variables
        :param random_walk: whether to slice in a random d-dimensional direction or hold all params const while slicing
        :param from_file: path to a .pkl file to load a saved slice from disk
        :returns: `fig, ax` with `num_slice` by `num_qoi` subplots
        """
        # Manage loading important quantities from file (if provided)
        xs, ys_model, ys_surr = None, None, None
        if from_file is not None:
            with open(Path(from_file), 'rb') as fd:
                slice_data = pickle.load(fd)
                slice_idx = slice_data['slice_idx']     # Must use same input slices as save file
                show_model = slice_data['show_model']   # Must use same model data as save file
                qoi_idx = slice_data['qoi_idx'] if qoi_idx is None else qoi_idx
                xs = slice_data['xs']
                model_dir = None  # Don't run or save any models if loading from file

        # Set default values (take up to the first 3 slices by default)
        rand_id = ''.join(random.choices(string.ascii_uppercase + string.digits, k=4))
        if model_dir is not None:
            os.mkdir(Path(model_dir) / f'sweep_{rand_id}')
        if nominal is None:
            nominal = dict()
        slice_idx = list(np.arange(0, min(3, len(self.exo_vars)))) if slice_idx is None else slice_idx
        qoi_idx = list(np.arange(0, min(3, len(self.coupling_vars)))) if qoi_idx is None else qoi_idx
        if isinstance(slice_idx[0], str | BaseRV):
            slice_idx = [self.exo_vars.index(var) for var in slice_idx]
        if isinstance(qoi_idx[0], str | BaseRV):
            qoi_idx = [self.coupling_vars.index(var) for var in qoi_idx]

        exo_bds = [var.bounds() for var in self.exo_vars]
        xlabels = [self.exo_vars[idx].to_tex(units=True) for idx in slice_idx]
        ylabels = [self.coupling_vars[idx].to_tex(units=True) for idx in qoi_idx]

        # Construct slice model inputs (if not provided)
        if xs is None:
            xs = np.zeros((N, len(slice_idx), len(self.exo_vars)))
            for i in range(len(slice_idx)):
                if random_walk:
                    # Make a random straight-line walk across d-cube
                    r0 = np.squeeze(self.sample_inputs((1,), use_pdf=False), axis=0)
                    r0[slice_idx[i]] = exo_bds[slice_idx[i]][0]             # Start slice at this lower bound
                    rf = np.squeeze(self.sample_inputs((1,), use_pdf=False), axis=0)
                    rf[slice_idx[i]] = exo_bds[slice_idx[i]][1]             # Slice up to this upper bound
                    xs[0, i, :] = r0
                    for k in range(1, N):
                        xs[k, i, :] = xs[k-1, i, :] + (rf-r0)/(N-1)
                else:
                    # Otherwise, only slice one variable
                    for j in range(len(self.exo_vars)):
                        if j == slice_idx[i]:
                            xs[:, i, j] = np.linspace(exo_bds[slice_idx[i]][0], exo_bds[slice_idx[i]][1], N)
                        else:
                            xs[:, i, j] = nominal.get(self.exo_vars[j], self.exo_vars[j].nominal)

        # Walk through each model that is requested by show_model
        if show_model is not None:
            if from_file is not None:
                ys_model = slice_data['ys_model']
            else:
                ys_model = list()
                for model in show_model:
                    output_dir = None
                    if model_dir is not None:
                        output_dir = (Path(model_dir) / f'sweep_{rand_id}' /
                                      str(model).replace('{', '').replace('}', '').replace(':', '=').replace("'", ''))
                        os.mkdir(output_dir)
                    ys_model.append(self(xs, use_model=model, model_dir=output_dir))
        if show_surr:
            ys_surr = self(xs) if from_file is None else slice_data['ys_surr']

        # Make len(qoi) by len(inputs) grid of subplots
        fig, axs = plt.subplots(len(qoi_idx), len(slice_idx), sharex='col', sharey='row')
        for i in range(len(qoi_idx)):
            for j in range(len(slice_idx)):
                if len(qoi_idx) == 1:
                    ax = axs if len(slice_idx) == 1 else axs[j]
                elif len(slice_idx) == 1:
                    ax = axs if len(qoi_idx) == 1 else axs[i]
                else:
                    ax = axs[i, j]
                x = xs[:, j, slice_idx[j]]
                if show_model is not None:
                    c = np.array([[0, 0, 0, 1], [0.5, 0.5, 0.5, 1]]) if len(show_model) <= 2 else (
                        plt.get_cmap('jet')(np.linspace(0, 1, len(show_model))))
                    for k in range(len(show_model)):
                        model_str = str(show_model[k]).replace('{', '').replace('}', '').replace(':', '=').replace("'", '')
                        model_ret = ys_model[k]
                        y_model = model_ret[:, j, qoi_idx[i]]
                        label = {'best': 'High-fidelity' if len(show_model) > 1 else 'Model',
                                 'worst': 'Low-fidelity'}.get(model_str, model_str)
                        ax.plot(x, y_model, ls='-', c=c[k, :], label=label)
                if show_surr:
                    y_surr = ys_surr[:, j, qoi_idx[i]]
                    ax.plot(x, y_surr, '--r', label='Surrogate')
                ylabel = ylabels[i] if j == 0 else ''
                xlabel = xlabels[j] if i == len(qoi_idx) - 1 else ''
                legend = (i == 0 and j == len(slice_idx) - 1)
                ax_default(ax, xlabel, ylabel, legend=legend)
        fig.set_size_inches(3 * len(slice_idx), 3 * len(qoi_idx))
        fig.tight_layout()

        # Save results (unless we were already loading from a save file)
        if from_file is None and self.root_dir is not None:
            fname = f's{",".join([str(i) for i in slice_idx])}_q{",".join([str(i) for i in qoi_idx])}'
            fname = f'sweep_rand{rand_id}_' + fname if random_walk else f'sweep_nom{rand_id}_' + fname
            fdir = Path(self.root_dir) if model_dir is None else Path(model_dir) / f'sweep_{rand_id}'
            fig.savefig(fdir / f'{fname}.png', dpi=300, format='png')
            save_dict = {'slice_idx': slice_idx, 'qoi_idx': qoi_idx, 'show_model': show_model, 'show_surr': show_surr,
                         'nominal': nominal, 'random_walk': random_walk, 'xs': xs, 'ys_model': ys_model,
                         'ys_surr': ys_surr}
            with open(fdir / f'{fname}.pkl', 'wb') as fd:
                pickle.dump(save_dict, fd)

        return fig, axs

    def plot_allocation(self, cmap: str = 'Blues', text_bar_width: float = 0.06, arrow_bar_width: float = 0.02):
        """Plot bar charts showing cost allocation during training.

        !!! Warning "Beta feature"
            This has pretty good default settings, but it might look terrible for your use. Mostly provided here as
            a template for making cost allocation bar charts. Please feel free to copy and edit in your own code.

        :param cmap: the colormap string identifier for `plt`
        :param text_bar_width: the minimum total cost fraction above which a bar will print centered model fidelity text
        :param arrow_bar_width: the minimum total cost fraction above which a bar will try to print text with an arrow;
                                below this amount, the bar is too skinny and won't print any text
        :returns: `fig, ax`, Figure and Axes objects
        """
        # Get total cost (including offline overhead)
        train_alloc, offline_alloc, cost_cum = self.get_allocation()
        total_cost = cost_cum[-1]
        for node, alpha_dict in offline_alloc.items():
            for alpha, cost in alpha_dict.items():
                total_cost += cost[1]

        # Remove nodes with cost=0 from alloc dicts (i.e. analytical models)
        remove_nodes = []
        for node, alpha_dict in train_alloc.items():
            if len(alpha_dict) == 0:
                remove_nodes.append(node)
        for node in remove_nodes:
            del train_alloc[node]
            del offline_alloc[node]

        # Bar chart showing cost allocation breakdown for MF system at end
        fig, axs = plt.subplots(1, 2, sharey='row')
        width = 0.7
        x = np.arange(len(train_alloc))
        xlabels = list(train_alloc.keys())
        cmap = plt.get_cmap(cmap)
        for k in range(2):
            ax = axs[k]
            alloc = train_alloc if k == 0 else offline_alloc
            ax.set_title('Online training' if k == 0 else 'Overhead')
            for j, (node, alpha_dict) in enumerate(alloc.items()):
                bottom = 0
                c_intervals = np.linspace(0, 1, len(alpha_dict))
                bars = [(alpha, cost, cost[1] / total_cost) for alpha, cost in alpha_dict.items()]
                bars = sorted(bars, key=lambda ele: ele[2], reverse=True)
                for i, (alpha, cost, frac) in enumerate(bars):
                    p = ax.bar(x[j], frac, width, color=cmap(c_intervals[i]), linewidth=1,
                               edgecolor=[0, 0, 0], bottom=bottom)
                    bottom += frac
                    if frac > text_bar_width:
                        ax.bar_label(p, labels=[f'{alpha}, {round(cost[0])}'], label_type='center')
                    elif frac > arrow_bar_width:
                        xy = (x[j] + width / 2, bottom - frac / 2)  # Label smaller bars with a text off to the side
                        ax.annotate(f'{alpha}, {round(cost[0])}', xy, xytext=(xy[0] + 0.2, xy[1]),
                                    arrowprops={'arrowstyle': '->', 'linewidth': 1})
                    else:
                        pass  # Don't label really small bars
            ax_default(ax, '', "Fraction of total cost" if k == 0 else '', legend=False)
            ax.set_xticks(x, xlabels)
            ax.set_xlim(left=-1, right=x[-1] + 1)
        fig.set_size_inches(2.5*len(x), 4)
        fig.tight_layout()

        if self.root_dir is not None:
            fig.savefig(Path(self.root_dir) / 'mf_allocation.png', dpi=300, format='png')

        return fig, axs

    def get_component(self, comp_name: str) -> ComponentSurrogate:
        """Return the `ComponentSurrogate` object for this component.

        :param comp_name: name of the component to return
        :returns: the `ComponentSurrogate` object
        """
        return self.graph.nodes[comp_name]['surrogate']

    def _print_title_str(self, title_str: str):
        """Log an important message."""
        self.logger.info('-' * int(len(title_str)/2) + title_str + '-' * int(len(title_str)/2))

    def save_to_file(self, filename: str, save_dir: str | Path = None):
        """Save the SystemSurrogate object to a .pkl file.

        :param filename: filename of the .pkl file to save to
        :param save_dir: overrides existing surrogate root directory if provided
        """
        if self.root_dir is None and save_dir is None:
            # Can't save to file if root_dir is None
            return

        save_dir = save_dir if save_dir is not None else str(Path(self.root_dir) / 'sys')
        if not Path(save_dir).is_dir():
            save_dir = '.'

        exec_temp = self.executor   # Temporarily save executor obj (can't pickle it)
        self.set_executor(None)
        with open(Path(save_dir) / filename, 'wb') as dill_file:
            dill.dump(self, dill_file)
        self.set_executor(exec_temp)
        self.logger.info(f'SystemSurrogate saved to {(Path(save_dir) / filename).resolve()}')

    def _set_output_dir(self, set_dict: dict[str: str | Path]):
        """Set the output directory for each component in `set_dict`.

        :param set_dict: a `dict` of component names (`str`) to their new output directories
        """
        for node, node_obj in self.graph.nodes.items():
            if node in set_dict:
                node_obj['surrogate']._set_output_dir(set_dict.get(node))

    def set_root_directory(self, root_dir: str | Path, stdout: bool = True):
        """Set the root to a new directory, for example if you move to a new filesystem.

        :param root_dir: new root directory
        :param stdout: whether to connect the logger to console (default)
        """
        self.root_dir = str(Path(root_dir).resolve())
        log_file = None
        if not (Path(self.root_dir) / 'sys').is_dir():
            os.mkdir(Path(self.root_dir) / 'sys')
        if not (Path(self.root_dir) / 'components').is_dir():
            os.mkdir(Path(self.root_dir) / 'components')
        for f in os.listdir(self.root_dir):
            if f.endswith('.log'):
                log_file = str((Path(self.root_dir) / f).resolve())
                break
        if log_file is None:
            fname = datetime.datetime.now(tz=timezone.utc).isoformat().split('.')[0].replace(':', '.') + 'UTC_sys.log'
            log_file = str((Path(self.root_dir) / fname).resolve())

        # Setup the log file
        self.log_file = log_file
        self.logger = get_logger(self.__class__.__name__, log_file=log_file, stdout=stdout)

        # Update model output directories
        for node, node_obj in self.graph.nodes.items():
            surr = node_obj['surrogate']
            surr.logger = get_logger(surr.__class__.__name__, log_file=log_file, stdout=stdout)
            surr.log_file = self.log_file
            if surr.save_enabled():
                output_dir = str((Path(self.root_dir) / 'components' / node).resolve())
                if not Path(output_dir).is_dir():
                    os.mkdir(output_dir)
                surr._set_output_dir(output_dir)

    def __getitem__(self, component: str) -> ComponentSurrogate:
        """Convenience method to get the `ComponentSurrogate object` from the `SystemSurrogate`.

        :param component: the name of the component to get
        :returns: the `ComponentSurrogate` object
        """
        return self.get_component(component)

    def __repr__(self):
        s = f'----SystemSurrogate----\nAdjacency: \n{nx.to_numpy_array(self.graph, dtype=int)}\n' \
            f'Exogenous inputs: {[str(var) for var in self.exo_vars]}\n'
        for node, node_obj in self.graph.nodes.items():
            s += f'Component: {node}\n{node_obj["surrogate"]}'
        return s

    def __str__(self):
        return self.__repr__()

    def set_executor(self, executor: Executor | None):
        """Set a new `concurrent.futures.Executor` object for parallel calls.

        :param executor: the new `Executor` object
        """
        self.executor = executor
        for node, node_obj in self.graph.nodes.items():
            node_obj['surrogate'].executor = executor

    @staticmethod
    def load_from_file(filename: str | Path, root_dir: str | Path = None, executor: Executor = None):
        """Load a `SystemSurrogate object` from file.

        :param filename: the .pkl file to load
        :param root_dir: folder to use as the root directory, (uses file's second parent directory by default)
        :param executor: a `concurrent.futures.Executor` object to set; clears it if None
        :returns: the `SystemSurrogate` object
        """
        if root_dir is None:
            root_dir = Path(filename).parent.parent     # Assume root/sys/filename.pkl

        with open(Path(filename), 'rb') as dill_file:
            sys_surr = dill.load(dill_file)
            sys_surr.set_executor(executor)
            sys_surr.set_root_directory(root_dir)
            sys_surr.logger.info(f'SystemSurrogate loaded from {Path(filename).resolve()}')

        return sys_surr

    @staticmethod
    def _constrained_lls(A: np.ndarray, b: np.ndarray, C: np.ndarray, d: np.ndarray) -> np.ndarray:
        """Minimize $||Ax-b||_2$, subject to $Cx=d$, i.e. constrained linear least squares.

        !!! Note
            See http://www.seas.ucla.edu/~vandenbe/133A/lectures/cls.pdf for more detail.

        :param A: `(..., M, N)`, vandermonde matrix
        :param b: `(..., M, 1)`, data
        :param C: `(..., P, N)`, constraint operator
        :param d: `(..., P, 1)`, constraint condition
        :returns: `(..., N, 1)`, the solution parameter vector `x`
        """
        M = A.shape[-2]
        dims = len(A.shape[:-2])
        T_axes = tuple(np.arange(0, dims)) + (-1, -2)
        Q, R = np.linalg.qr(np.concatenate((A, C), axis=-2))
        Q1 = Q[..., :M, :]
        Q2 = Q[..., M:, :]
        Q1_T = np.transpose(Q1, axes=T_axes)
        Q2_T = np.transpose(Q2, axes=T_axes)
        Qtilde, Rtilde = np.linalg.qr(Q2_T)
        Qtilde_T = np.transpose(Qtilde, axes=T_axes)
        Rtilde_T_inv = np.linalg.pinv(np.transpose(Rtilde, axes=T_axes))
        w = np.linalg.pinv(Rtilde) @ (Qtilde_T @ Q1_T @ b - Rtilde_T_inv @ d)

        return np.linalg.pinv(R) @ (Q1_T @ b - Q2_T @ w)



"""`utils.py`

Provides some basic utilities for the package.

Includes
--------
- `load_variables`: convenience function for loading RVs from a .json config file
- `get_logger`: logging utility with nice formatting
- `ax_default`: plotting utility with nice formatting
- `approx_hess`: finite difference approximation of the Hessian
- `batch_normal_sample`: helper function to sample from arbitrarily-sized Gaussian distribution(s)
- `ndscatter`: plotting utility for n-dimensional data
"""
import json
from pathlib import Path
import logging
import sys

import matplotlib.pyplot as plt
import matplotlib
from matplotlib.pyplot import cycler
from matplotlib.colors import LinearSegmentedColormap, ListedColormap
from matplotlib.ticker import StrMethodFormatter, AutoLocator, FuncFormatter
import scipy.stats as st
import numpy as np
from numpy.linalg.linalg import LinAlgError

from amisc.rv import BaseRV, UniformRV, NormalRV, ScalarRV


LOG_FORMATTER = logging.Formatter("%(asctime)s \u2014 [%(levelname)s] \u2014 %(name)-20s \u2014 %(message)s")


def load_variables(variables: list[str], file: Path | str) -> list[BaseRV]:
    """Load a list of BaseRV objects from a variables json `file`.

    :param variables: a list of str ids for variables to find in `file`
    :param file: json file to search for variable definitions
    :returns rvs: a list of corresponding `BaseRV` objects
    """
    with open(Path(file), 'r') as fd:
        data = json.load(fd)

    rvs = []
    keys = ['id', 'tex', 'description', 'units', 'param_type', 'nominal', 'domain']
    for str_id in variables:
        if str_id in data:
            var_info = data.get(str_id)
            kwargs = {key: var_info.get(key) for key in keys if var_info.get(key)}
            match var_info.get('rv_type', 'none'):
                case 'uniform_bds':
                    bds = var_info.get('rv_params')
                    rvs.append(UniformRV(bds[0], bds[1], **kwargs))
                case 'uniform_pct':
                    rvs.append(UniformRV(var_info.get('rv_params'), 'pct', **kwargs))
                case 'uniform_tol':
                    rvs.append(UniformRV(var_info.get('rv_params'), 'tol', **kwargs))
                case 'normal':
                    mu, std = var_info.get('rv_params')
                    rvs.append(NormalRV(mu, std, **kwargs))
                case 'none':
                    # Make a plain stand-in scalar RV object (no uncertainty)
                    rvs.append(ScalarRV(**kwargs))
                case other:
                    raise NotImplementedError(f'RV type "{var_info.get("rv_type")}" is not known.')
        else:
            raise ValueError(f'You have requested the variable {str_id}, but it was not found in {file}. '
                             f'Please add a definition of {str_id} to {file} or construct it on your own.')

    return rvs


def get_logger(name: str, stdout=True, log_file: str | Path = None) -> logging.Logger:
    """Return a file/stdout logger with the given name.

    :param name: the name of the logger to return
    :param stdout: whether to add a stdout handler to the logger
    :param log_file: add file logging to this file (optional)
    :returns: the logger
    """
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    logger.handlers.clear()
    if stdout:
        std_handler = logging.StreamHandler(sys.stdout)
        std_handler.setFormatter(LOG_FORMATTER)
        logger.addHandler(std_handler)
    if log_file is not None:
        f_handler = logging.FileHandler(log_file, mode='a', encoding='utf-8')
        f_handler.setLevel(logging.DEBUG)
        f_handler.setFormatter(LOG_FORMATTER)
        logger.addHandler(f_handler)

    return logger


def approx_hess(func: callable, theta: np.ndarray, pert=0.01) -> np.ndarray:
    """Approximate Hessian of `func` at a specified `theta` location.

    :param func: expects to be called as `func(theta) -> (..., y_dim)`, where `y_dim=1` (scalar funcs only)
    :param theta: `(..., theta_dim)`, points to linearize model about
    :param pert: perturbation for approximate partial derivatives
    :returns H: `(..., theta_dim, theta_dim)`, the approximate Hessian `(theta_dim, theta_dim)` at all locations (...)
    """
    theta = np.atleast_1d(theta)
    shape = theta.shape[:-1]                # (*)
    theta_dim = theta.shape[-1]             # Number of parameters
    dtheta = pert * theta

    # Return the Hessians (..., theta_dim, theta_dim)
    H = np.empty(shape + (theta_dim, theta_dim))

    for i in range(theta_dim):
        for j in range(i, theta_dim):
            # Allocate space at 4 grid points (n1=-1, p1=+1)
            theta_n1_n1 = np.copy(theta)
            theta_p1_p1 = np.copy(theta)
            theta_n1_p1 = np.copy(theta)
            theta_p1_n1 = np.copy(theta)

            # Perturbations to theta in each direction
            theta_n1_n1[..., i] -= dtheta[..., i]
            theta_n1_n1[..., j] -= dtheta[..., j]
            f_n1_n1 = func(theta_n1_n1)

            theta_p1_p1[..., i] += dtheta[..., i]
            theta_p1_p1[..., j] += dtheta[..., j]
            f_p1_p1 = func(theta_p1_p1)

            theta_n1_p1[..., i] -= dtheta[..., i]
            theta_n1_p1[..., j] += dtheta[..., j]
            f_n1_p1 = func(theta_n1_p1)

            theta_p1_n1[..., i] += dtheta[..., i]
            theta_p1_n1[..., j] -= dtheta[..., j]
            f_p1_n1 = func(theta_p1_n1)

            res = (f_n1_n1 + f_p1_p1 - f_n1_p1 - f_p1_n1) / np.expand_dims(4 * dtheta[..., i] * dtheta[..., j],
                                                                           axis=-1)

            # Hessian only computed for scalar functions, y_dim=1 on last axis
            H[..., i, j] = np.squeeze(res, axis=-1)
            H[..., j, i] = np.squeeze(res, axis=-1)

    return H


def batch_normal_sample(mean: np.ndarray | float, cov: np.ndarray | float, size: tuple | int = ()) -> np.ndarray:
    """Batch sample multivariate normal distributions (pretty much however you want).

    :param mean: `(..., dim)`, expected values, where dim is the random variable dimension
    :param cov: `(..., dim, dim)`, covariance matrices
    :param size: shape of additional samples
    :returns samples: `(*size, ..., dim)`, samples from multivariate distributions
    """
    mean = np.atleast_1d(mean)
    cov = np.atleast_2d(cov)

    if isinstance(size, int):
        size = (size, )
    shape = size + np.broadcast_shapes(mean.shape, cov.shape[:-1])
    x_normal = np.random.standard_normal((*shape, 1)).astype(np.float32)
    samples = np.squeeze(np.linalg.cholesky(cov) @ x_normal, axis=-1) + mean
    return samples


def ax_default(ax: plt.Axes, xlabel='', ylabel='', legend=True, cmap='tab10'):
    """Nice default formatting for plotting X-Y data.

    :param ax: the axes to apply these settings to
    :param xlabel: the xlabel to set for `ax`
    :param ylabel: the ylabel to set for `ax`
    :param legend: whether to show a legend
    :param cmap: colormap to use for cycling
    """
    plt.rcParams["axes.prop_cycle"] = get_cycle(cmap)
    plt.rc('xtick', labelsize='small')
    plt.rc('ytick', labelsize='small')
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.tick_params(axis='x', direction='in')
    ax.tick_params(axis='y', direction='in')
    if legend:
        leg = ax.legend(fancybox=True)
        frame = leg.get_frame()
        frame.set_edgecolor('k')


def get_cycle(cmap: str | matplotlib.colors.Colormap, num_colors: int = None):
    """Get a color cycler for plotting.

    :param cmap: a string specifier of a matplotlib colormap (or a colormap instance)
    :param num_colors: the number of colors to cycle through
    """
    use_index = False
    if isinstance(cmap, str):
        use_index = cmap in ['Pastel1', 'Pastel2', 'Paired', 'Accent', 'Dark2', 'Set1', 'Set2', 'Set3',
                             'tab10', 'tab20', 'tab20b', 'tab20c']
        cmap = matplotlib.cm.get_cmap(cmap)
    if num_colors is None:
        num_colors = cmap.N
    if cmap.N > 100:
        use_index = False
    elif isinstance(cmap, LinearSegmentedColormap):
        use_index = False
    elif isinstance(cmap, ListedColormap):
        use_index = True
    if use_index:
        ind = np.arange(int(num_colors)) % cmap.N
        return cycler("color", cmap(ind))
    else:
        colors = cmap(np.linspace(0, 1, num_colors))
        return cycler("color", colors)


def ndscatter(samples: np.ndarray, labels: list[str] = None, tick_fmts: list[str] = None, plot='scatter',
              cmap='viridis', bins=20, z: np.ndarray = None, cb_label=None, cb_norm='linear', subplot_size=3):
    """Triangle scatter plots of n-dimensional samples.

    !!! Warning
        Best for `dim < 10`. You can shrink the `subplot_size` to assist graphics loading time.

    :param samples: `(N, dim)` samples to plot
    :param labels: list of axis labels of length `dim`
    :param tick_fmts: list of str.format() specifiers for ticks, e.g `['{x: ^10.2f}', ...]`, of length `dim`
    :param plot: 'hist' for 2d hist plot, 'kde' for kernel density estimation, or 'scatter' (default)
    :param cmap: the matplotlib string specifier of a colormap
    :param bins: number of bins in each dimension for histogram marginals
    :param z: `(N,)` a performance metric corresponding to `samples`, used to color code the scatter plot if provided
    :param cb_label: label for color bar (if `z` is provided)
    :param cb_norm: `str` or `plt.colors.Normalize`, normalization method for plotting `z` on scatter plot
    :param subplot_size: size in inches of a single 2d marginal subplot
    :returns fig, axs: the `plt` Figure and Axes objects, (returns an additional `cb_fig, cb_ax` if `z` is specified)
    """
    N, dim = samples.shape
    x_min = np.min(samples, axis=0)
    x_max = np.max(samples, axis=0)
    if labels is None:
        labels = [f"x{i}" for i in range(dim)]
    if z is None:
        z = np.zeros(N)
    if cb_label is None:
        cb_label = 'Performance metric'

    def tick_format_func(value, pos):
        if value > 1:
            return f'{value:.2f}'
        if value > 0.01:
            return f'{value:.4f}'
        if value < 0.01:
            return f'{value:.2E}'
    default_ticks = FuncFormatter(tick_format_func)
    # if tick_fmts is None:
    #     tick_fmts = ['{x:.2G}' for i in range(dim)]

    # Set up triangle plot formatting
    fig, axs = plt.subplots(dim, dim, sharex='col', sharey='row')
    for i in range(dim):
        for j in range(dim):
            ax = axs[i, j]
            if i == j:                      # 1d marginals on diagonal
                # ax.get_shared_y_axes().remove(ax)
                ax._shared_axes['y'].remove(ax)
                ax.spines['top'].set_visible(False)
                ax.spines['right'].set_visible(False)
                ax.spines['left'].set_visible(False)
                if i == 0:
                    ax.get_yaxis().set_ticks([])
            if j > i:                       # Clear the upper triangle
                ax.axis('off')
            if i == dim - 1:                # Bottom row
                ax.set_xlabel(labels[j])
                ax.xaxis.set_major_locator(AutoLocator())
                formatter = StrMethodFormatter(tick_fmts[j]) if tick_fmts is not None else default_ticks
                ax.xaxis.set_major_formatter(formatter)
            if j == 0 and i > 0:            # Left column
                ax.set_ylabel(labels[i])
                ax.yaxis.set_major_locator(AutoLocator())
                formatter = StrMethodFormatter(tick_fmts[i]) if tick_fmts is not None else default_ticks
                ax.yaxis.set_major_formatter(formatter)

    # Plot marginals
    for i in range(dim):
        for j in range(dim):
            ax = axs[i, j]
            if i == j:                      # 1d marginals (on diagonal)
                c = plt.get_cmap(cmap)(0)
                if plot == 'kde':
                    kernel = st.gaussian_kde(samples[:, i])
                    x = np.linspace(x_min[i], x_max[i], 1000)
                    ax.fill_between(x, y1=kernel(x), y2=0, lw=0, alpha=0.3, facecolor=c)
                    ax.plot(x, kernel(x), ls='-', c=c, lw=1.5)
                else:
                    ax.hist(samples[:, i], edgecolor='black', color=c, density=True, alpha=0.5,
                            linewidth=1.2, bins='auto')
            if j < i:                       # 2d marginals (lower triangle)
                ax.set_xlim([x_min[j], x_max[j]])
                ax.set_ylim([x_min[i], x_max[i]])
                if plot == 'scatter':
                    sc = ax.scatter(samples[:, j], samples[:, i], s=1.5, c=z, cmap=cmap, norm=cb_norm)
                elif plot == 'hist':
                    ax.hist2d(samples[:, j], samples[:, i], bins=bins, density=True, cmap=cmap)
                elif plot == 'kde':
                    kernel = st.gaussian_kde(samples[:, [j, i]].T)
                    xg, yg = np.meshgrid(np.linspace(x_min[j], x_max[j], 60), np.linspace(x_min[i], x_max[i], 60))
                    x = np.vstack([xg.ravel(), yg.ravel()])
                    zg = np.reshape(kernel(x), xg.shape)
                    ax.contourf(xg, yg, zg, 5, cmap=cmap, alpha=0.9)
                    ax.contour(xg, yg, zg, 5, colors='k', linewidths=1.5)
                else:
                    raise NotImplementedError('This plot type is not known. plot=["hist", "kde", "scatter"]')

    fig.set_size_inches(subplot_size * dim, subplot_size * dim)
    fig.tight_layout()

    # Plot colorbar in standalone figure
    if np.max(z) > 0 and plot == 'scatter':
        cb_fig, cb_ax = plt.subplots(figsize=(1.5, 6))
        cb_fig.subplots_adjust(right=0.7)
        cb_fig.colorbar(sc, cax=cb_ax, orientation='vertical', label=cb_label)
        cb_fig.tight_layout()
        return fig, axs, cb_fig, cb_ax

    return fig, axs



"""Adaptive multi-index stochastic collocation for metamodeling/surrogates of multidisciplinary systems.

- Author - Joshua Eckels (eckelsjd@umich.edu)
- License - GNU GPLv3
"""
import numpy as np

from amisc.interpolator import BaseInterpolator
from amisc.rv import BaseRV

__version__ = "0.1.1"

# Custom types that are used frequently
IndexSet = list[tuple[tuple, tuple]]
MiscTree = dict[str: dict[str: float | BaseInterpolator]]
InterpResults = BaseInterpolator | tuple[list[int | tuple | str], np.ndarray, BaseInterpolator]
IndicesRV = list[int | str | BaseRV] | int | str | BaseRV



import numpy as np
import matplotlib.pyplot as plt

from amisc.interpolator import LagrangeInterpolator
from amisc.rv import UniformRV
from amisc.utils import ax_default
from amisc.examples.models import tanh_func, nonlinear_wave


def test_tensor_product_1d(plots=False):
    beta = (3,)
    x_var = UniformRV(0, 1)
    x_grid = np.linspace(0, 1, 100).reshape((100, 1))
    y_grid = tanh_func(x_grid)['y']
    interp = LagrangeInterpolator(beta, x_var, model=tanh_func)
    interp.set_yi()
    y_interp = interp(x_grid)

    # Refine
    beta2 = (4,)
    interp2 = interp.refine(beta2)
    y2_interp = interp2(x_grid)

    # Compute errors
    N = 1000
    xtest = np.random.rand(N, 1)
    ytest = interp2(xtest)
    ytruth = tanh_func(xtest)['y']
    l2_error = np.sqrt(np.mean((ytest - ytruth) ** 2)) / np.sqrt(np.mean(ytruth ** 2))
    assert l2_error < 1e-1

    # Plot results
    if plots:
        fig, ax = plt.subplots()
        ax.plot(x_grid, y_grid, '-k', label='Model')
        ax.plot(interp.xi, interp.yi, 'or', markersize=6, label=r'Training data')
        ax.plot(x_grid, y_interp, '-r', label=r'$\beta=3$')
        ax.plot(interp2.xi, interp2.yi, 'ob', markersize=4)
        ax.plot(x_grid, y2_interp, '-b', label=r'$\beta=4$')
        ax_default(ax, r'Input', r'Output', legend=True)
        fig.tight_layout()
        plt.show()


def test_tensor_product_2d(plots=False):
    bb_2d_func = lambda x: nonlinear_wave(x, env_var=0.2**2, wave_amp=0.3)
    beta = (3, 3)
    x_vars = [UniformRV(0, 1), UniformRV(0, 1)]
    N = 50
    x_grid = np.linspace(0, 1, N)
    xg, yg = np.meshgrid(x_grid, x_grid)
    xg = xg.reshape((N, N, 1))
    yg = yg.reshape((N, N, 1))
    x = np.concatenate((xg, yg), axis=-1)
    z = bb_2d_func(x)['y']

    # Set up interpolator
    interp = LagrangeInterpolator(beta, x_vars, model=bb_2d_func)
    interp.set_yi()
    z_interp = interp(x)
    error = np.abs(z_interp - z)

    # Refine interpolator
    beta2 = (3, 4)
    interp2 = interp.refine(beta2)
    z2_interp = interp2(x)
    error2 = np.abs(z2_interp - z)
    vmin = min(np.min(z_interp), np.min(z), np.min(z2_interp))
    vmax = max(np.max(z_interp), np.max(z), np.max(z2_interp))
    emin = min(np.min(error), np.min(error2))
    emax = max(np.max(error), np.max(error2))

    # Compute errors
    N = 1000
    xtest = np.random.rand(N, 2)
    ytest = interp2(xtest)
    ytruth = bb_2d_func(xtest)['y']
    l2_error = np.sqrt(np.mean((ytest - ytruth)**2)) / np.sqrt(np.mean(ytruth**2))
    assert l2_error < 1e-1

    if plots:
        fig, ax = plt.subplots(2, 3)
        c1 = ax[0, 0].contourf(xg.squeeze(), yg.squeeze(), z.squeeze(), 60, cmap='coolwarm', vmin=vmin, vmax=vmax)
        plt.colorbar(c1, ax=ax[0, 0])
        ax[0, 0].set_title('True function')
        ax_default(ax[0, 0], r'$x_1$', r'$x_2$', legend=False)
        c2 = ax[0, 1].contourf(xg.squeeze(), yg.squeeze(), z_interp.squeeze(), 60, cmap='coolwarm', vmin=vmin, vmax=vmax)
        ax[0, 1].plot(interp.xi[:, 0], interp.xi[:, 1], 'o', markersize=6, markerfacecolor='green')
        plt.colorbar(c2, ax=ax[0, 1])
        ax[0, 1].set_title('Interpolant')
        ax_default(ax[0, 1], r'$x_1$', '', legend=False)
        c3 = ax[0, 2].contourf(xg.squeeze(), yg.squeeze(), error.squeeze(), 60, cmap='viridis', vmin=emin, vmax=emax)
        ax[0, 2].plot(interp.xi[:, 0], interp.xi[:, 1], 'o', markersize=6, markerfacecolor='green')
        plt.colorbar(c3, ax=ax[0, 2])
        ax[0, 2].set_title('Absolute error')
        ax_default(ax[0, 2], r'$x_1$', '', legend=False)
        c1 = ax[1, 0].contourf(xg.squeeze(), yg.squeeze(), z.squeeze(), 60, cmap='coolwarm', vmin=vmin, vmax=vmax)
        plt.colorbar(c1, ax=ax[1, 0])
        ax[1, 0].set_title('True function')
        ax_default(ax[1, 0], r'$x_1$', r'$x_2$', legend=False)
        c2 = ax[1, 1].contourf(xg.squeeze(), yg.squeeze(), z2_interp.squeeze(), 60, cmap='coolwarm', vmin=vmin, vmax=vmax)
        ax[1, 1].plot(interp2.xi[:, 0], interp2.xi[:, 1], 'o', markersize=6, markerfacecolor='green')
        plt.colorbar(c2, ax=ax[1, 1])
        ax[1, 1].set_title('Refined')
        ax_default(ax[1, 1], r'$x_1$', '', legend=False)
        c3 = ax[1, 2].contourf(xg.squeeze(), yg.squeeze(), error2.squeeze(), 60, cmap='viridis', vmin=emin, vmax=emax)
        ax[1, 2].plot(interp2.xi[:, 0], interp2.xi[:, 1], 'o', markersize=6, markerfacecolor='green')
        plt.colorbar(c3, ax=ax[1, 2])
        ax[1, 2].set_title('Absolute error')
        ax_default(ax[1, 2], r'$x_1$', '', legend=False)
        fig.set_size_inches(15, 11)
        fig.tight_layout()
        plt.show()
